#!/usr/bin/env python

import argparse
import os
import platform
import re
import string
import sys
from enum import Enum
from typing import Optional

from py010parser import c_ast, parse_file

from fandango.converters.FandangoConverter import FandangoConverter


class BitfieldOrder(Enum):
    LeftToRight = 0
    RightToLeft = 1


class Endianness(Enum):
    LittleEndian = 0
    BigEndian = 1


class BTFandangoConverterVisitor(c_ast.NodeVisitor):
    def __init__(
        self,
        *,
        start_symbol: str = "<start>",
        use_regexes: bool = False,
        bitfield_order: BitfieldOrder = BitfieldOrder.LeftToRight,
        endianness: Endianness = Endianness.BigEndian,
    ):
        self.bitfield_order: BitfieldOrder = bitfield_order
        self.endianness: Endianness = endianness
        self.start_symbol = start_symbol
        self.use_regexes = use_regexes

        self.defs: dict[str, str] = {}
        self.forced_elems: list[str] = []
        self.forced_complements: list[str] = []
        self.forced_vars: list[str] = []
        self.seen: set[str] = set()
        self.context: list[str] = []
        self.constraints: list[str] = []
        self.not_handled: list[str] = []
        self.vars: dict[Optional[str], str] = {}
        self.renames: dict[str, str] = {}
        self.in_code = False

    def cond(self):
        return " and ".join(self.context)

    def _char(self, c):
        if c in string.printable:
            return c
        return f"\\x{ord(c):02x}"

    def char(self, c):
        if c in string.printable:
            return "b" + repr(c)
        return f"b'{self._char(c)}'"

    def not_char(self, c):
        if self.use_regexes:
            return f"br'[^{self._char(c)}]'"
        chars = []
        for i in range(256):
            if chr(i) != c:
                chars.append(self.char(chr(i)))
        return "(" + " | ".join(chars) + ")" + "  # not " + self.char(c)

    def not_short(self, value):
        low_byte = self._char(chr(ord(value) % 256))
        high_byte = self._char(chr(ord(value) // 256))
        return f"(/[^{low_byte}]./ | /[{low_byte}][^{high_byte}]/)"

    def spec(self, symbol: Optional[str] = None, indent=0) -> str:
        if symbol is None:
            symbol = self.start_symbol
        if indent == 0:
            self.seen = set()
        if symbol not in self.defs or symbol in self.seen:
            return ""
        self.seen.add(symbol)  # avoid infinite recursion

        expansion = self.defs[symbol]
        s = "  " * indent + f"{symbol} ::= {expansion}\n"
        nonterminals = re.findall(r"(<[^>]+>)", expansion)
        for nonterminal in nonterminals:
            s += self.spec(nonterminal, indent + 1)

        if indent == 0:
            if self.constraints:
                s += "\n"
            for constraint in self.constraints:
                s += f"where {constraint}\n"
            if self.not_handled:
                s += "\n"
            for not_handled in self.not_handled:
                s += f"# FIXME: {not_handled}\n"
        return s

    def visit(self, node):
        # print("Visiting", node.__class__.__name__)
        method_name = "visit_" + node.__class__.__name__
        visitor = getattr(self, method_name, self.generic_visit)
        return visitor(node)

    def add_def(self, base_name, members):
        for rename in self.renames.keys():
            members = members.replace(rename, self.renames[rename])

        i = 1
        name = base_name
        while name in self.defs:
            name = f"<{base_name[1:-1]}_{i}>"
            i += 1

        if name != base_name:  # record the rename
            self.renames[base_name] = name

        # print(f"Adding {name} ::= {members}")
        self.defs[name] = members

    def add_constraint(self, constraint):
        for rename in self.renames.keys():
            constraint = constraint.replace(rename, self.renames[rename])
        self.constraints.append(constraint)

    def generic_visit(self, node) -> str:
        print("Ignoring", node.__class__.__name__)
        for _, child in node.children():
            self.visit(child)
        return ""

    def generic_join(self, node, sep: str = " ") -> str:
        s = ""
        for _, child in node.children():
            member = self.visit(child)
            if s and member:
                s += sep
            if member:
                s += member
        return s

    def visit_ID(self, node: c_ast.ID) -> str:
        return node.name

    def visit_BinaryOp(self, node: c_ast.BinaryOp) -> str:
        left = self.visit(node.left)
        right = self.visit(node.right)
        return f"{left} {node.op} {right}"

    def visit_UnaryOp(self, node: c_ast.UnaryOp) -> str:
        expr = self.visit(node.expr)
        return f"{node.op}{expr}"

    def visit_ExprList(self, node: c_ast.ExprList) -> str:
        return self.generic_join(node)

    def visit_TypeDecl(self, node: c_ast.TypeDecl) -> str:
        return self.generic_join(node)

    def visit_FuncCall(self, node: c_ast.FuncCall) -> str:
        if self.in_code:
            return ""

        self.in_code = True
        name = self.visit(node.name)
        self.in_code = False

        if name == "BitfieldLeftToRight":
            self.bitfield_order = BitfieldOrder.LeftToRight
            return ""
        if name == "BitfieldRightToLeft":
            self.bitfield_order = BitfieldOrder.RightToLeft
            return ""

        if name == "LittleEndian":
            self.endianness = Endianness.LittleEndian
            return ""
        if name == "BigEndian":
            self.endianness = Endianness.BigEndian
            return ""

        if name == "RequiresVersion":
            return ""
        if name == "Warning":
            return ""

        self.not_handled.append(f"{name}()")
        return ""

    def visit_FuncDef(self, node: c_ast.FuncDef) -> str:
        if self.in_code:
            return ""

        self.in_code = True
        decl = self.visit(node.decl)
        self.not_handled.append(f"{decl}()")
        self.in_code = False
        return ""

    def visit_Compound(self, node: c_ast.Compound) -> str:
        return self.generic_join(node)

    def visit_StructRef(self, node: c_ast.StructRef) -> str:
        name = self.visit(node.name)
        field = self.visit(node.field)
        return f"<{name}>.<{field}>"

    def visit_FileAST(self, node: c_ast.FileAST) -> str:
        members = self.generic_join(node)
        self.add_def(self.start_symbol, members)
        return ""

    def visit_EmptyStatement(self, node: c_ast.EmptyStatement) -> str:
        return ""

    def visit_Typedef(self, node: c_ast.Typedef) -> str:
        members = self.generic_join(node)
        if node.name:
            self.add_def(f"<{node.name}>", members)
        return ""

    def visit_Struct(self, node: c_ast.Struct) -> str:
        members = []
        is_bitfield = True
        for _, child in node.children():
            if hasattr(child, "bitsize") and child.bitsize:
                bitsize = eval(self.visit(child.bitsize))
                if bitsize > 0:
                    continue

            is_bitfield = False
            break

        for _, child in node.children():
            value = None
            var: Optional[str] = None
            if self.forced_elems and self.forced_complements:
                value = self.forced_elems[0]
                self.forced_elems = self.forced_elems[1:]
                self.forced_complements = self.forced_complements[1:]
            elif self.forced_elems:
                value = self.forced_elems[0]
                self.forced_elems = self.forced_elems[1:]
            elif self.forced_complements:
                value = self.forced_complements[0]
                self.forced_complements = self.forced_complements[1:]

            if self.forced_vars:
                var = self.forced_vars[0]
                self.forced_vars = self.forced_vars[1:]

            elem = self.visit(child)
            if value:
                self.add_def(f"<{child.name}>", value)
                self.vars[var] = f"<{child.name}>"

            if elem:
                members.append(elem)

        if is_bitfield and self.bitfield_order == BitfieldOrder.RightToLeft:
            members.reverse()

        members_str = " ".join(members)

        # if is_bitfield:
        #     print(f"Bitfield {node.name} = {members_str}")
        # else:
        #     print(f"Struct {node.name} = {members_str}")

        if node.name:
            self.add_def(f"<{node.name}>", members_str)
            return f"<{node.name}>"

        return members_str

    def visit_Decl(self, node: c_ast.Decl):
        if self.in_code:
            return node.name

        if "local" in node.quals:
            # treat like an assignment
            if node.init:
                init = self.visit(node.init)
                self.vars[node.name] = node.init
                self.not_handled.append(f"{node.name} = {init}")
            return ""

        if node.bitsize:
            bitsize = eval(self.visit(node.bitsize))
            if bitsize == 1:
                m = "<bit>"
            else:
                m = f"<bit>{{{bitsize}}}"
        else:
            type_ = node.type
            if isinstance(type_, c_ast.ArrayDecl):
                array_decl: c_ast.ArrayDecl = type_
                type_name = self.visit(array_decl.type)
                dim = self.visit(array_decl.dim)
                if dim in self.vars and isinstance(self.vars[dim], str):
                    symbol = self.vars[dim]
                    constraint = f"len(<{node.name}>) == ord(str({symbol}))"
                    m = f"{type_name}*  # {constraint}; see below"
                    self.add_constraint(constraint)
                else:
                    try:
                        dim_n = int(dim)
                        m = f"{type_name}{{{dim_n}}}"
                    except ValueError:
                        m = f"{type_name}*  # FIXME: must be {{{dim}}}"
            else:
                m = self.visit(type_)

        self.add_def(f"<{node.name}>", m)
        return f"<{node.name}>"

    def visit_ArrayDecl(self, node: c_ast.ArrayDecl) -> str:
        type_ = self.visit(node.type)
        dim = self.visit(node.dim)
        try:
            dim_n = int(dim)
            return f"{type_}{{{dim_n}}}"
        except ValueError:
            pass
        return f"{type_}*  # FIXME: must be {{{dim}}}"

    def visit_Constant(self, node: c_ast.Constant) -> str:
        return f"{node.value}"

    def visit_Return(self, node: c_ast.Return) -> str:
        # We assume an early return, i.e. return -1
        self.add_constraint(f"not ({self.cond()})")
        return ""

    def visit_IdentifierType(self, node: c_ast.IdentifierType) -> str:
        name = "_".join(node.names)
        if self.in_code:
            return name
        else:
            return f"<{name}>"

    def visit_Assignment(self, node: c_ast.Assignment) -> str:
        lvalue = self.visit(node.lvalue)
        rvalue = self.visit(node.rvalue)
        self.vars[lvalue] = node.rvalue
        self.not_handled.append(f"{lvalue} = {rvalue}")
        return ""

    def force_elems(self, cond, iftrue=True):
        # Convert lookaheads into expected bytes
        # as in `if (ReadUShort(FTell()) == 0x0121) ...`
        if not isinstance(cond, c_ast.BinaryOp):
            return

        binary_op: c_ast.BinaryOp = cond
        complement = (
            binary_op.op == "==" and not iftrue or binary_op.op == "!=" and iftrue
        )

        left = binary_op.left
        right = binary_op.right
        var = None

        # Replace variable names by the last recorded expression
        if isinstance(left, c_ast.ID):
            left = self.vars.get(binary_op.left.name)
            if left is None:
                return
            var = binary_op.left.name
        elif isinstance(binary_op.right, c_ast.ID):
            right = self.vars.get(binary_op.right.name)
            if right is None:
                return
            var = binary_op.right.name

        # Identify the function call
        funccall: Optional[c_ast.FuncCall] = None
        if isinstance(left, c_ast.FuncCall):
            funccall = left
        elif isinstance(right, c_ast.FuncCall):
            funccall = right

        if not funccall:
            return

        # Identify the constant value
        value = None
        if isinstance(left, c_ast.Constant):
            value = eval(left.value)
        elif isinstance(right, c_ast.Constant):
            value = eval(right.value)

        if value is None:
            return
        if isinstance(value, int):
            value = chr(value)

        func = funccall.name
        if not isinstance(func, c_ast.ID):
            return

        name = func.name
        if name == "ReadUByte":
            if complement:
                self.forced_complements += [self.not_char(value)]
            else:
                self.forced_elems += [self.char(value)]
            if var:
                self.forced_vars += [var]

        elif name == "ReadUShort":
            if self.endianness == Endianness.BigEndian:
                next_bytes = [chr(ord(value) // 256), chr(ord(value) % 256)]
            else:  # LittleEndian
                next_bytes = [chr(ord(value) % 256), chr(ord(value) // 256)]

            if complement:
                # Too complex at this point; should be something like
                # self.forced_complements += [not_short(value)]
                # Here's a simpler alternative, only checking for one byte
                self.forced_complements += [self.not_char(next_bytes[0])]
            else:
                self.forced_elems += [
                    self.char(next_bytes[0]),
                    self.char(next_bytes[1]),
                ]

    def visit_While(self, node: c_ast.While | c_ast.For) -> str:
        if self.in_code:
            return ""

        cond = self.visit(node.cond)

        self.force_elems(node.cond, iftrue=True)
        self.context.append(cond)
        body = self.visit(node.stmt)
        self.context.pop()

        self.force_elems(node.cond, iftrue=False)
        if not body:
            return ""
        return f"({body})*"

    def visit_For(self, node: c_ast.For) -> str:
        return self.visit_While(node)

    def visit_If(self, node: c_ast.If) -> str:
        if self.in_code:
            return ""

        self.force_elems(node.cond, iftrue=True)

        cond = self.visit(node.cond)
        self.context.append(cond)
        iftrue = self.visit(node.iftrue)
        self.context.pop()

        if len(node.children()) <= 2:
            if not iftrue:
                return ""
            return f"{iftrue}?"

        self.force_elems(node.cond, iftrue=False)

        self.context.append("not " + cond)
        iffalse = self.visit(node.iffalse)
        self.context.pop()

        if iftrue and iffalse:
            return f"{iftrue} | {iffalse}"
        if iftrue:
            return iftrue
        if iffalse:
            return iffalse
        return ""


class BTFandangoConverter(FandangoConverter):
    """Convert a binary template to a Fandango specification."""

    def __init__(self, filename: str):
        super().__init__(filename)

        try:
            if platform.system() == "Darwin":
                ast = parse_file(filename, cpp_args="-xc++")
            else:
                ast = parse_file(filename)
            self.ast = ast
        finally:
            try:
                # The 010 parser leaves behind some temporary files
                os.remove("lextab.py")
                os.remove("yacctab.py")
            except FileNotFoundError:
                pass

    def to_fan(
        self,
        start_symbol: str = "<start>",
        use_regexes: bool = True,
        bitfield_order: BitfieldOrder = BitfieldOrder.LeftToRight,
        endianness: Endianness = Endianness.BigEndian,
    ) -> str:
        visitor = BTFandangoConverterVisitor(
            start_symbol=start_symbol,
            use_regexes=use_regexes,
            bitfield_order=bitfield_order,
            endianness=endianness,
        )
        visitor.visit(self.ast)
        header = f"""# Automatically generated from {self.filename!r}.
#
"""
        return header + visitor.spec()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert a binary template to a Fandango specification"
    )
    parser.add_argument(
        "--no-regexes",
        action="store_false",
        dest="use_regexes",
        default=True,
        help="do not use regexes",
    )
    parser.add_argument(
        "--endianness", choices=["little", "big"], help="set endianness"
    )
    parser.add_argument(
        "--bitfield-order",
        choices=["left-to-right", "right-to-left"],
        help="set bitfield order",
    )
    parser.add_argument(
        "filename", nargs="+", type=str, help=".bt binary template file"
    )

    args = parser.parse_args(sys.argv[1:])
    if args.endianness == "little":
        endianness = Endianness.LittleEndian
    else:
        endianness = Endianness.BigEndian
    if args.bitfield_order == "left-to-right":
        bitfield_order = BitfieldOrder.LeftToRight
    else:
        bitfield_order = BitfieldOrder.RightToLeft

    for filename in args.filename:
        converter = BTFandangoConverter(filename)
        fan = converter.to_fan(
            start_symbol="<start>",
            use_regexes=args.use_regexes,
            bitfield_order=bitfield_order,
            endianness=endianness,
        )
        print(fan)



import enum
from typing import Any, Optional
from fandango.constraints.base import GeneticBase
from fandango.language.tree import DerivationTree


class Comparison(enum.Enum):
    """
    Enum class for comparison operations.
    """

    EQUAL = "=="
    NOT_EQUAL = "!="
    GREATER = ">"
    GREATER_EQUAL = ">="
    LESS = "<"
    LESS_EQUAL = "<="


class ComparisonSide(enum.Enum):
    """
    Enum class for comparison side.
    """

    LEFT = "left"
    RIGHT = "right"


class FailingTree:
    """
    Class to represent a failing tree, i.e., a tree that does not satisfy a given constraint.
    """

    def __init__(
        self,
        tree: DerivationTree,
        cause: "GeneticBase",
        suggestions: Optional[list[tuple[Comparison, Any, ComparisonSide]]] = None,
    ):
        """
        Initialize the FailingTree with the given tree, cause, and suggestions.

        :param DerivationTree tree: The tree that failed to satisfy the constraint.
        :param GeneticBase cause: The cause of the failure.
        :param Optional[list[tuple[Comparison, Any, ComparisonSide]]] suggestions: The list of suggestions to
        which causes the cause to fail.
        """
        self.tree = tree
        self.cause = cause
        self.suggestions = suggestions or []

    def __hash__(self):
        return hash((self.tree, self.cause))

    def __eq__(self, other):
        return self.tree == other.tree and self.cause == other.cause

    def __repr__(self):
        return f"FailingTree({self.tree}, {self.cause}, {self.suggestions})"

    def __str__(self):
        return self.__repr__()


class BoundsFailingTree(FailingTree):
    """
    Class to represent a failing tree that has was generated by a bounds constraint.
    """

    def __init__(
        self,
        tree: DerivationTree,
        starting_rep_tree: DerivationTree,
        ending_rep_tree: DerivationTree,
        starting_rep_value: DerivationTree,
        ending_rep_value: DerivationTree,
        cause: "GeneticBase",
        suggestions: Optional[list[tuple[Comparison, Any, ComparisonSide]]] = None,
    ):
        """
        Initialize the BoundsFailingTree with the given tree, cause, lower bound, upper bound, and suggestions.

        :param DerivationTree tree: The tree that failed to satisfy the constraint.
        :param DerivationTree starting_rep_tree: The first tree generated by the repetition.
        :param DerivationTree ending_rep_tree: The last tree generated by the repetition.
        :param DerivationTree starting_rep_value: The reference to the subtree giving min repetition value
        :param DerivationTree ending_rep_value: A reference to the subtree giving max repetition value
        :param GeneticBase cause: The cause of the failure.
        :param Optional[list[tuple[Comparison, Any, ComparisonSide]]] suggestions: The list of suggestions to
        which causes the cause to fail.
        """
        super().__init__(tree, cause, suggestions)
        self.starting_rep_tree = starting_rep_tree
        self.ending_rep_tree = ending_rep_tree
        self.starting_rep_value = starting_rep_value
        self.ending_rep_value = ending_rep_value

    def __repr__(self):
        return f"BoundsFailingTree({self.tree}, {self.cause}, {str(self.starting_rep_value)}, {self.ending_rep_value}, {self.suggestions})"



from abc import ABC, abstractmethod
from collections.abc import Callable, Generator, Iterable
import itertools
import logging
import time
from typing import IO, Optional
from fandango.constraints.constraint import Constraint
from fandango.constraints.soft import SoftValue
from fandango.language.grammar import FuzzingMode, ParsingMode
from fandango.language.grammar.grammar import Grammar
from fandango.language.parse import parse
from fandango.language.tree import DerivationTree
from fandango.logger import LOGGER
from fandango.evolution.algorithm import Fandango as FandangoStrategy
from fandango.errors import FandangoFailedError, FandangoParseError

DEFAULT_MAX_GENERATIONS = 500


class FandangoBase(ABC):
    """Public Fandango API"""

    # The parser to be used
    parser = "auto"  # 'auto', 'cpp', 'python', or 'legacy'

    def __init__(
        self,
        fan_files: str | IO | list[str | IO],
        constraints: Optional[list[str]] = None,
        *,
        logging_level: Optional[int] = None,
        use_cache: bool = True,
        use_stdlib: bool = True,
        lazy: bool = False,
        start_symbol: Optional[str] = None,
        includes: Optional[list[str]] = None,
    ):
        """
        Initialize a Fandango object.
        :param fan_files: One (open) .fan file, one string, or a list of these
        :param constraints: List of constraints (as strings); default: []
        :param use_cache: If True (default), cache parsing results
        :param use_stdlib: If True (default), use the standard library
        :param lazy: If True, the constraints are evaluated lazily
        :param start_symbol: The grammar start symbol (default: "<start>")
        :param includes: A list of directories to search for include files
        """
        self._start_symbol = start_symbol if start_symbol is not None else "<start>"
        LOGGER.setLevel(logging_level if logging_level is not None else logging.WARNING)
        grammar, self._constraints = parse(
            fan_files,
            constraints,
            use_cache=use_cache,
            use_stdlib=use_stdlib,
            lazy=lazy,
            start_symbol=start_symbol,
            includes=includes,
        )
        if grammar is None:
            raise FandangoParseError(
                position=0,
                message="Failed to parse grammar, Grammar is None",
            )
        self._grammar = grammar

    @property
    def grammar(self):
        return self._grammar

    @grammar.setter
    def grammar(self, value):
        self._grammar = value

    @property
    def constraints(self):
        return self._constraints

    @constraints.setter
    def constraints(self, value):
        self._constraints = value

    @property
    def start_symbol(self):
        return self._start_symbol

    @start_symbol.setter
    def start_symbol(self, value):
        self._start_symbol = value

    @property
    def logging_level(self):
        return LOGGER.getEffectiveLevel()

    @logging_level.setter
    def logging_level(self, value):
        LOGGER.setLevel(value)

    @abstractmethod
    def init_population(
        self, *, extra_constraints: Optional[list[str]] = None, **settings
    ) -> None:
        """
        Initialize a Fandango population.
        :param extra_constraints: Additional constraints to apply
        :param settings: Additional settings for the evolution algorithm
        :return: A list of derivation trees
        """
        pass

    @abstractmethod
    def generate_solutions(
        self,
        max_generations: Optional[int] = None,
        mode: FuzzingMode = FuzzingMode.COMPLETE,
    ) -> Generator[DerivationTree, None, None]:
        """
        Generate trees that conform to the language.

        Will initialize a population with default settings if none has been initialized before.
        Initialization can be done manually with `init_population` for more flexibility.

        :param max_generations: Maximum number of generations to evolve through
        :return: A generator for solutions to the language
        """
        pass

    @abstractmethod
    def fuzz(
        self,
        *,
        extra_constraints: Optional[list[str]] = None,
        solution_callback: Callable[[DerivationTree, int], None] = lambda _a, _b: None,
        desired_solutions: Optional[int] = None,
        max_generations: Optional[int] = None,
        infinite: bool = False,
        mode: FuzzingMode = FuzzingMode.COMPLETE,
        **settings,
    ) -> list[DerivationTree]:
        """
        Create a Fandango population.
        :param extra_constraints: Additional constraints to apply
        :param solution_callback: What to do with each solution; receives the solution and a unique index
        :param settings: Additional settings for the evolution algorithm
        :return: A list of derivation trees
        """
        pass

    @abstractmethod
    def parse(
        self, word: str | bytes | DerivationTree, *, prefix: bool = False, **settings
    ) -> Generator[Optional[DerivationTree], None, None]:
        """
        Parse a string according to spec.
        :param word: The string to parse
        :param prefix: If True, allow incomplete parsing
        :param settings: Additional settings for the parse function
        :return: A generator of derivation trees
        """
        pass


class Fandango(FandangoBase):
    """Evolutionary testing with Fandango."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fandango = None

    @classmethod
    def _with_parsed(
        cls,
        grammar: Grammar,
        constraints: list[Constraint | SoftValue],
        *,
        start_symbol: Optional[str] = None,
        logging_level: Optional[int] = None,
    ) -> "FandangoBase":
        LOGGER.setLevel(logging_level if logging_level is not None else logging.WARNING)
        obj = cls.__new__(cls)  # bypass __init__ to prevent the need for double parsing
        obj._grammar = grammar
        obj._constraints = constraints
        obj.fandango = None
        obj._start_symbol = start_symbol if start_symbol is not None else "<start>"
        return obj

    def init_population(
        self, *, extra_constraints: Optional[list[str]] = None, **settings
    ) -> None:
        """
        Initialize a Fandango population.
        :param extra_constraints: Additional constraints to apply
        :param settings: Additional settings for the evolution algorithm
        :return: A list of derivation trees
        """
        LOGGER.info("---------- Initializing base population ----------")

        start_symbol = settings.pop("start_symbol", self._start_symbol)

        constraints = self.constraints[:]
        if extra_constraints:
            _, extra_constraints_parsed = parse(
                [],
                extra_constraints,
                given_grammars=[self.grammar],
                start_symbol=start_symbol,
            )
            constraints += extra_constraints_parsed

        self.fandango = FandangoStrategy(
            self.grammar, constraints, start_symbol=start_symbol, **settings
        )
        LOGGER.info("---------- Done initializing base population ----------")

    def generate_solutions(
        self,
        max_generations: Optional[int] = None,
        mode: FuzzingMode = FuzzingMode.COMPLETE,
    ) -> Generator[DerivationTree, None, None]:
        """
        Generate trees that conform to the language.

        Will initialize a population with default settings if none has been initialized before.
        Initialization can be done manually with `init_population` for more flexibility.

        :param max_generations: Maximum number of generations to evolve through
        :return: A generator for solutions to the language
        """
        if self.fandango is None:
            self.init_population()
            assert self.fandango is not None

        LOGGER.info(
            f"---------- Generating {'' if max_generations is None else f' for {max_generations} generations'}----------"
        )
        start_time = time.time()
        yield from self.fandango.generate(max_generations=max_generations, mode=mode)
        LOGGER.info(
            f"---------- Done generating {'' if max_generations is None else f' for {max_generations} generations'}----------"
        )
        LOGGER.info(f"Time taken: {(time.time() - start_time):.2f} seconds")

    def fuzz(
        self,
        *,
        extra_constraints: Optional[list[str]] = None,
        solution_callback: Callable[[DerivationTree, int], None] = lambda _a, _b: None,
        desired_solutions: Optional[int] = None,
        max_generations: Optional[int] = None,
        infinite: bool = False,
        mode: FuzzingMode = FuzzingMode.COMPLETE,
        **settings,
    ) -> list[DerivationTree]:
        """
        Create a Fandango population.
        :param extra_constraints: Additional constraints to apply
        :param solution_callback: What to do with each solution; receives the solution and a unique index
        :param settings: Additional settings for the evolution algorithm
        :return: A list of derivation trees
        """

        # force-(re-)initialize if settings changed
        if extra_constraints is not None or settings is not None:
            self.init_population(extra_constraints=extra_constraints, **settings)
        assert self.fandango is not None

        if mode == FuzzingMode.IO:
            match desired_solutions:
                case None:
                    LOGGER.warning(
                        "Fandango IO will only return a single solution for now, manually set with -n 1 to hide this warning"
                    )
                case 1:
                    pass
                case _:
                    LOGGER.warning(
                        "Fandango IO only supports desired-solution values of 1 for now, overriding value"
                    )
            desired_solutions = 1

        if max_generations is None and desired_solutions is None and not infinite:
            LOGGER.info(
                f"Infinite is not set and neither max_generations nor desired_solutions are specified. Limiting to default max_generations of {DEFAULT_MAX_GENERATIONS}"
            )
            max_generations = DEFAULT_MAX_GENERATIONS
        else:
            LOGGER.debug(
                f"Limiting fuzzing to max_generations: {max_generations} and desired_solutions: {desired_solutions}"
            )

        if infinite:
            if max_generations is not None:
                LOGGER.warn("Infinite mode is activated, overriding max_generations")
            max_generations = None  # infinite overrides max_generations

        generator: Iterable[DerivationTree] = self.generate_solutions(
            max_generations=max_generations, mode=mode
        )
        if desired_solutions is not None:
            LOGGER.info(f"Generating {desired_solutions} solutions")
            generator = itertools.islice(generator, desired_solutions)

        solutions = []
        for i, s in enumerate(generator):
            solutions.append(s)
            solution_callback(s, i)

        if desired_solutions is not None and len(solutions) < desired_solutions:
            warnings_are_errors = settings.get("warnings_are_errors", False)
            best_effort = settings.get("best_effort", False)
            if (
                self.fandango.average_population_fitness
                < self.fandango.evaluator.expected_fitness
            ):
                LOGGER.error("Population did not converge to a perfect population")
                if warnings_are_errors:
                    raise FandangoFailedError("Failed to find a perfect solution")
                elif best_effort:
                    return self.fandango.population

            LOGGER.error(
                f"Only found {len(solutions)} perfect solutions, instead of the required {desired_solutions}"
            )
            if warnings_are_errors:
                raise FandangoFailedError(
                    "Failed to find the required number of perfect solutions"
                )
            elif best_effort:
                return self.fandango.population[:desired_solutions]

        return solutions

    def parse(
        self, word: str | bytes | DerivationTree, *, prefix: bool = False, **settings
    ) -> Generator[Optional[DerivationTree], None, None]:
        """
        Parse a string according to spec.
        :param word: The string to parse
        :param prefix: If True, allow incomplete parsing
        :param settings: Additional settings for the parse function
        :return: A generator of derivation trees
        """
        if prefix:
            mode = ParsingMode.INCOMPLETE
        else:
            mode = ParsingMode.COMPLETE

        tree_generator = self.grammar.parse_forest(
            word, mode=mode, start=self._start_symbol, **settings
        )
        try:
            peek = next(tree_generator)
        except StopIteration:
            peek = None

        if peek is None:
            position = self.grammar.max_position() + 1
            raise FandangoParseError(position=position)

        self.grammar.populate_sources(peek)
        yield peek
        yield from tree_generator



import random
import struct
from abc import ABC, abstractmethod
from enum import Enum, auto

from fandango import DerivationTree

# Magic Values taken from LibAFL
ARITH_MAX = 35
MAX_SIZE = 1048576

# Interesting values from LibAFL
INTERESTING_VALUES = {
    1: [-128, -1, 0, 1, 16, 32, 64, 100, 127],
    2: [
        -128,
        -1,
        0,
        1,
        16,
        32,
        64,
        100,
        127,
        -32768,
        -129,
        128,
        255,
        256,
        512,
        1000,
        1024,
        4096,
        32767,
    ],
    4: [
        -128,
        -1,
        0,
        1,
        16,
        32,
        64,
        100,
        127,
        -32768,
        -129,
        128,
        255,
        256,
        512,
        1000,
        1024,
        4096,
        32767,
        -2147483648,
        -100663046,
        -32769,
        32768,
        65535,
        65536,
        100663045,
        2147483647,
    ],
}


def _wrapping_add(a: int, b: int, byte_size: int = 1) -> int:
    return (a + b) % (256**byte_size)


def _wrapping_sub(a: int, b: int, byte_size: int = 1) -> int:
    return (a - b) % (256**byte_size)


def _get_random_index(input: bytearray) -> int:
    """
    Returns a random index in the input_.

    :param input: The input_ to get the index from.
    :return: A random index in the input_, always valid as an index for the input_.
    """
    return random.randint(0, len(input) - 1)


def _get_random_range_of_length(input: bytearray, length: int) -> tuple[int, int]:
    """
    Returns a random range of the given length in the input_.

    :param input: The input_ to get the range from.
    :param length: The length of the range to get.
    :return: A tuple of the start (inclusive) and end (exclusive) of the range.
    """

    if length == 0:
        raise ValueError("Length cannot be 0")
    if length > len(input):
        raise ValueError("Length is greater than the input_ size")
    start = random.randint(0, len(input) - length)
    end = start + length
    return start, end


def _get_random_range_with_max_length(
    input: bytearray, max_length: int
) -> tuple[int, int]:
    """
    Returns a random range of a length between 1 and the max length in the input_.

    :param input: The input_ to get the range from.
    :param max_length: The maximum length of the range to get.
    :return: A tuple of the start (inclusive) and end (exclusive) of the range.
    """
    if max_length > len(input):
        raise ValueError("Max length is greater than the input_ size")
    length = random.randint(1, max_length)
    return _get_random_range_of_length(input, length)


class ByteLevelMutationOperator(ABC):
    """Base class for all mutation operators that perform byte-level mutations on input_ data."""

    @abstractmethod
    def mutate(
        self,
        input: bytearray,
    ) -> bool:
        """
        Abstract method to perform byte-level mutation.

        :param input: The input_ to mutate.
        :return: True if the input_ was mutated, False otherwise.
        """
        pass


class BitFlipMutation(ByteLevelMutationOperator):
    """Flips a random bit in a random byte of the input_."""

    def mutate(self, input: bytearray) -> bool:
        if len(input) == 0:
            return False

        offset = _get_random_index(input)
        input[offset] ^= 1 << random.randint(0, 7)
        return True


class ByteFlipMutation(ByteLevelMutationOperator):
    """Flips all bits in a random byte of the input_."""

    def mutate(self, input: bytearray) -> bool:
        if len(input) == 0:
            return False

        offset = _get_random_index(input)
        input[offset] ^= 0xFF
        return True


class ByteIncMutation(ByteLevelMutationOperator):
    """Increments a random byte in the input_ by 1."""

    def mutate(self, input: bytearray) -> bool:
        if len(input) == 0:
            return False

        offset = _get_random_index(input)
        input[offset] = _wrapping_add(input[offset], 1)
        return True


class ByteDecMutation(ByteLevelMutationOperator):
    """Decrements a random byte in the input_ by 1."""

    def mutate(self, input: bytearray) -> bool:
        if len(input) == 0:
            return False

        offset = _get_random_index(input)
        input[offset] = _wrapping_sub(input[offset], 1)
        return True


class ByteNegMutation(ByteLevelMutationOperator):
    """Negates a random byte in the input_."""

    def mutate(self, input: bytearray) -> bool:
        if len(input) == 0:
            return False

        offset = _get_random_index(input)
        value = _wrapping_add(input[offset] ^ 0xFF, 1)
        if value == input[offset]:
            return False
        input[offset] = value
        return True


class ByteRandMutation(ByteLevelMutationOperator):
    """Replaces a random byte in the input_ with a random value."""

    def mutate(self, input: bytearray) -> bool:
        if len(input) == 0:
            return False

        offset = _get_random_index(input)
        new_val = random.randint(0, 255)
        if new_val == input[offset]:
            return False
        input[offset] = new_val
        return True


class ArithmeticOperation(Enum):
    """Enum for arithmetic operations used in mutations."""

    ADD = auto()
    SUB = auto()


def _get_format_string(num_bytes: int, little_endian: bool) -> str:
    if num_bytes == 1:
        format_string_size = "B"
    elif num_bytes == 2:
        format_string_size = "H"
    elif num_bytes == 4:
        format_string_size = "I"
    elif num_bytes == 8:
        format_string_size = "Q"
    else:
        raise ValueError(f"Unsupported number of bytes: {num_bytes}")

    format_string_endianness = "<" if little_endian else ">"
    return format_string_endianness + format_string_size


class MultiByteArithmeticMutation(ByteLevelMutationOperator):
    """Performs an addition or subtraction on a random (multi-byte) value in the input_, in either little or big endian."""

    def __init__(
        self, num_bytes: int, operation: ArithmeticOperation, little_endian: bool
    ):
        self.num_bytes = num_bytes
        self.operation = operation
        self.little_endian = little_endian
        self.format_string = _get_format_string(num_bytes, little_endian)
        match operation:
            case ArithmeticOperation.ADD:
                self.operation_function = _wrapping_add
            case ArithmeticOperation.SUB:
                self.operation_function = _wrapping_sub

    def mutate(self, input: bytearray) -> bool:
        if len(input) < self.num_bytes:
            return False

        offset, end = _get_random_range_of_length(input, self.num_bytes)
        val = struct.unpack(self.format_string, input[offset:end])[0]
        new_val = self.operation_function(
            val, random.randint(2, ARITH_MAX + 1), byte_size=self.num_bytes
        )
        # ensure a positive value for packing
        new_val = new_val % 256**self.num_bytes
        if new_val == val:
            return False
        struct.pack_into(self.format_string, input, offset, new_val)
        return True

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(num_bytes={self.num_bytes}, "
            f"operation={self.operation}, "
            f"little_endian={self.little_endian})"
        )


class MultiByteInterestingMutation(ByteLevelMutationOperator):
    """Replaces a random (multi-byte) value in the input_ with a potentially interesting value."""

    def __init__(self, num_bytes: int):
        self.num_bytes = num_bytes
        self.format_string = _get_format_string(num_bytes, False)

    def mutate(self, input: bytearray) -> bool:
        if len(input) < self.num_bytes:
            return False

        start, end = _get_random_range_of_length(input, self.num_bytes)
        replacement_value = (
            random.choice(INTERESTING_VALUES[self.num_bytes]) % 256**self.num_bytes
        )
        if struct.unpack(self.format_string, input[start:end])[0] == replacement_value:
            return False
        struct.pack_into(self.format_string, input, start, replacement_value)
        return True

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(num_bytes={self.num_bytes})"


class BytesDeleteMutation(ByteLevelMutationOperator):
    """Deletes random bytes from the input_."""

    def mutate(self, input: bytearray) -> bool:
        if len(input) < 2:
            return False

        num_bytes = random.randint(1, len(input) - 1)
        offset, end = _get_random_range_of_length(input, num_bytes)
        del input[offset:end]
        return True


class BytesExpandMutation(ByteLevelMutationOperator):
    """Extends the input_ by a subslice of itself."""

    def mutate(self, input: bytearray) -> bool:
        size = len(input)
        if size >= MAX_SIZE or size == 0:
            return False

        # Don't let input_ grow beyond MAX_SIZE
        max_value = min(16, size, MAX_SIZE - size)
        if max_value <= 0:
            return False
        amount = random.randint(1, max_value)

        start, end = _get_random_range_of_length(input, amount)
        input.extend(input[start:end])
        return True


class BytesInsertMutation(ByteLevelMutationOperator):
    """Inserts a slice consisting of a random value from the input_ into a random position in the input_."""

    def mutate(self, input: bytearray) -> bool:
        size = len(input)
        if size >= MAX_SIZE or size == 0:
            return False

        value = random.choice(input)

        amount = random.randint(1, 16)
        if size + amount > MAX_SIZE:
            if MAX_SIZE > size:
                amount = MAX_SIZE - size
            else:
                return False

        offset = _get_random_index(input)
        input[offset:offset] = [value] * amount
        return True


class BytesRandInsertMutation(ByteLevelMutationOperator):
    """Inserts a slice consisting of a random value into a random position in the input_."""

    def mutate(self, input: bytearray) -> bool:
        size = len(input)
        if size >= MAX_SIZE or size == 0:
            return False

        value = random.randint(0, 255)

        amount = random.randint(1, 16)
        if size + amount > MAX_SIZE:
            if MAX_SIZE > size:
                amount = MAX_SIZE - size
            else:
                return False

        offset = _get_random_index(input)
        input[offset:offset] = [value] * amount
        return True


class BytesSetMutation(ByteLevelMutationOperator):
    """Sets a random range in the input_ to a random value from the input_."""

    def mutate(self, input: bytearray) -> bool:
        size = len(input)
        if size == 0:
            return False

        value = random.choice(input)
        start, end = _get_random_range_with_max_length(input, min(size, 16))
        new_val = bytearray([value] * (end - start))
        if input[start:end] == new_val:
            return False
        input[start:end] = new_val
        return True


class BytesRandSetMutation(ByteLevelMutationOperator):
    """Sets a random range in the input_ to a random value."""

    def mutate(self, input: bytearray) -> bool:
        if len(input) == 0:
            return False

        value = random.randint(0, 255)
        start, end = _get_random_range_with_max_length(input, len(input))
        new_val = bytearray([value] * (end - start))
        if input[start:end] == new_val:
            return False
        input[start:end] = new_val
        return True


class BytesCopyMutation(ByteLevelMutationOperator):
    """Copies a random range in the input_ to a random position in the input_."""

    def mutate(self, input: bytearray) -> bool:
        size = len(input)
        if size < 2:
            return False

        target_start = _get_random_index(input)
        source_start, source_end = _get_random_range_with_max_length(
            input, size - target_start
        )
        target_end = target_start + source_end - source_start

        if input[target_start:target_end] == input[source_start:source_end]:
            return False
        input[target_start:target_end] = input[source_start:source_end]
        return True


class BytesInsertCopyMutation(ByteLevelMutationOperator):
    """Inserts a copy of a random range in the input_ into a random position in the input_."""

    def mutate(self, input: bytearray) -> bool:
        size = len(input)
        if size < 2 or size >= MAX_SIZE:
            return False

        target_start, _ = _get_random_range_of_length(input, 2)
        max_insert_len = random.randint(
            1, min(16, size - target_start, MAX_SIZE - size)
        )
        source_start, _ = _get_random_range_of_length(input, max_insert_len)
        temp_buf = input[source_start : source_start + max_insert_len]
        input[target_start + max_insert_len : size + max_insert_len] = input[
            target_start:size
        ]
        input[target_start : target_start + max_insert_len] = temp_buf
        return True


class BytesSwapMutation(ByteLevelMutationOperator):
    """Swaps a random range in the input_ with a random range in the input_."""

    def mutate(self, input: bytearray) -> bool:
        size = len(input)
        if size < 2:
            return False
        first_start, first_end = _get_random_range_with_max_length(input, len(input))
        if first_start > 0 and random.choice([True, False]):
            # second range comes before first
            second_start, second_end = _get_random_range_with_max_length(
                bytearray(input[:first_start]), first_start
            )

            input_copy = input[second_start:first_end]

            second_temp = input[second_start:second_end]
            middle_temp = input[second_end:first_start]
            first_temp = input[first_start:first_end]

            input[second_start : second_start + len(first_temp)] = first_temp
            input[
                second_start
                + len(first_temp) : second_start
                + len(first_temp)
                + len(middle_temp)
            ] = middle_temp
            input[first_end - len(second_temp) : first_end] = second_temp

            if input[second_start:first_end] == input_copy:
                return False
        elif first_end < size - 1:
            # second range comes after first
            second_start, second_end = _get_random_range_with_max_length(
                bytearray(input[first_end:]), size - first_end
            )
            second_start, second_end = second_start + first_end, second_end + first_end

            input_copy = input[first_start:second_end]

            first_temp = input[first_start:first_end]
            middle_temp = input[first_end:second_start]
            second_temp = input[second_start:second_end]

            input[first_start : first_start + len(second_temp)] = second_temp
            input[
                first_start
                + len(second_temp) : first_start
                + len(second_temp)
                + len(middle_temp)
            ] = middle_temp
            input[second_end - len(first_temp) : second_end] = first_temp

            if input[first_start:second_end] == input_copy:
                return False
        else:
            return False

        return True


def havoc_mutations() -> tuple[ByteLevelMutationOperator, ...]:
    return (
        BitFlipMutation(),
        ByteFlipMutation(),
        ByteIncMutation(),
        ByteDecMutation(),
        ByteNegMutation(),
        ByteRandMutation(),
        MultiByteArithmeticMutation(1, ArithmeticOperation.ADD, False),
        MultiByteArithmeticMutation(1, ArithmeticOperation.SUB, False),
        # for 1 byte, little endian and big endian are the same
        MultiByteArithmeticMutation(2, ArithmeticOperation.ADD, False),
        MultiByteArithmeticMutation(2, ArithmeticOperation.SUB, False),
        MultiByteArithmeticMutation(2, ArithmeticOperation.ADD, True),
        MultiByteArithmeticMutation(2, ArithmeticOperation.SUB, True),
        MultiByteArithmeticMutation(4, ArithmeticOperation.ADD, False),
        MultiByteArithmeticMutation(4, ArithmeticOperation.SUB, False),
        MultiByteArithmeticMutation(4, ArithmeticOperation.ADD, True),
        MultiByteArithmeticMutation(4, ArithmeticOperation.SUB, True),
        MultiByteArithmeticMutation(8, ArithmeticOperation.ADD, False),
        MultiByteArithmeticMutation(8, ArithmeticOperation.SUB, False),
        MultiByteArithmeticMutation(8, ArithmeticOperation.ADD, True),
        MultiByteArithmeticMutation(8, ArithmeticOperation.SUB, True),
        MultiByteInterestingMutation(1),
        MultiByteInterestingMutation(2),
        MultiByteInterestingMutation(4),
        BytesDeleteMutation(),
        BytesExpandMutation(),
        BytesInsertMutation(),
        BytesRandInsertMutation(),
        BytesSetMutation(),
        BytesRandSetMutation(),
        BytesCopyMutation(),
        BytesInsertCopyMutation(),
        BytesSwapMutation(),
    )


def havoc_mutate(
    input_: DerivationTree,
    mutations: (
        list[ByteLevelMutationOperator] | tuple[ByteLevelMutationOperator, ...]
    ) = havoc_mutations(),
    max_stack_pow=7,
    nop_probability=0,
) -> bytes:
    """
    Mutates the input using the given mutations, defaulting to a .

    :param input_: The input to mutate.
    :param mutations: The mutations to use.
    :param max_stack_pow: The maximum power of 2 for the stack size.
    :return: The mutated input.
    """
    if random.random() < nop_probability:
        return input_.to_bytes()

    inp = bytearray(input_.to_bytes())
    for _ in range(1 << random.randint(0, max_stack_pow)):
        mutation = random.choice(mutations)
        mutation.mutate(inp)
    fixed = bytes(inp)
    return fixed



import random
import time
from typing import Optional

from fandango.errors import FandangoFailedError, FandangoParseError, FandangoValueError
from fandango.io import FandangoIO
from fandango.io.packetforecaster import PacketForecaster
from fandango.language import Grammar, NonTerminal, DerivationTree
from fandango.language.grammar import ParsingMode
from fandango.language.grammar.parser.iterative_parser import IterativeParser


def _find_next_fragment(
    role_sender: str, messages: list[tuple[str, str, str | bytes]], start_idx: int = 0
) -> tuple[int, Optional[str | bytes]]:
    """
    Find the next message fragment sent by the specified sender in the list of messages.
    Returns the index of the message and the message fragment.
    """
    for idx in range(start_idx, len(messages)):
        sender, recipient, msg_fragment = messages[idx]
        if sender == role_sender:
            return idx, msg_fragment
    return -1, None


def parse_next_remote_packet(
    grammar: Grammar,
    forecast: PacketForecaster.ForecastingResult,
    io_instance: FandangoIO,
):
    if len(io_instance.get_received_msgs()) == 0:
        return None, None

    # Wait till we receive a message from one of the parties in the forecast
    received_parties = list(map(lambda x: x[0], io_instance.get_received_msgs()))
    wait_for_msg_time = 10
    start_time = time.time()
    while not forecast.contains_any_party(received_parties):
        if time.time() - start_time > wait_for_msg_time:
            if len(received_parties) == 0:
                raise FandangoFailedError(
                    "Timeout while waiting for message. No message has been received."
                )
            else:
                raise FandangoValueError(
                    "Unexpected party sent message. Expected: "
                    + " | ".join(forecast.get_msg_parties())
                    + f". Received: {received_parties}."
                    + f" Messages: {io_instance.get_received_msgs()}"
                )
        time.sleep(0.025)
        received_parties = list(map(lambda x: x[0], io_instance.get_received_msgs()))

    msg_sender = None
    # We might have received messages from different parties. Select a party that sent a message and is
    # in the current forecast.
    for idx, (msg_sender, msg_recipient, _) in enumerate(
        io_instance.get_received_msgs()
    ):
        if msg_sender in forecast.get_msg_parties():
            break

    assert msg_sender is not None

    forecast_non_terminals = forecast[msg_sender]
    available_non_terminals = set(forecast_non_terminals.get_non_terminals())

    # Initialize parsers for each non-terminal in the forecast applicable for the sender
    nt_parsers: dict[NonTerminal, IterativeParser] = dict()
    for non_terminal in available_non_terminals:
        forecast_packet = forecast_non_terminals[non_terminal]
        hookin_data = random.choice(list(forecast_packet.paths))
        hookin_tree = hookin_data.tree
        assert hookin_tree is not None
        path = list(map(lambda x: x[0], filter(lambda x: not x[1], hookin_data.path)))
        hookin_point = hookin_tree.get_last_by_path(path)
        nt_parsers[non_terminal] = IterativeParser(grammar.rules)
        nt_parsers[non_terminal].new_parse(
            start=non_terminal, mode=ParsingMode.COMPLETE, hookin_parent=hookin_point
        )

    continue_parse = True
    complete_parses: dict[NonTerminal, tuple[int, DerivationTree]] = dict()
    current_fragment_idx: int = -1
    parameter_parsing_exception_tuple = None
    wait_for_completion_time = 1
    while continue_parse:

        # Find the next message fragment sent by the selected sender
        start_time = time.time()
        next_fragment_idx, next_fragment = _find_next_fragment(
            msg_sender, io_instance.get_received_msgs(), current_fragment_idx + 1
        )
        while next_fragment is None:
            next_fragment_idx, next_fragment = _find_next_fragment(
                msg_sender, io_instance.get_received_msgs(), current_fragment_idx + 1
            )
            if time.time() - start_time > wait_for_completion_time:
                if len(complete_parses) == 0:
                    raise FandangoFailedError(
                        "Timeout while waiting for next message fragment from "
                        + msg_sender
                        + ". Messages: "
                        + str(io_instance.get_received_msgs())
                    )
                else:
                    continue_parse = False
                    break
            time.sleep(0.025)
        if not continue_parse:
            break

        assert next_fragment is not None
        current_fragment_idx = next_fragment_idx

        for non_terminal in set(available_non_terminals):
            parser = nt_parsers[non_terminal]
            parse_tree = next(parser.consume(next_fragment), None)
            if parse_tree is not None:
                parse_tree = parser.collapse(parse_tree)
                assert parse_tree is not None
                forecast_packet = forecast_non_terminals[non_terminal]
                parse_tree.sender = forecast_packet.node.sender
                parse_tree.recipient = forecast_packet.node.recipient
                try:
                    grammar.populate_sources(parse_tree)
                    complete_parses[non_terminal] = (current_fragment_idx, parse_tree)
                except FandangoParseError as e:
                    parameter_parsing_exception_tuple = (non_terminal, e, parse_tree)
            if not parser.can_continue():
                available_non_terminals.remove(non_terminal)
        continue_parse = len(available_non_terminals) > 0

    if len(complete_parses) == 0:
        if parameter_parsing_exception_tuple is not None:
            applicable_nt, parameter_parsing_exception, complete_msg = (
                parameter_parsing_exception_tuple
            )
            raise FandangoFailedError(
                f"Couldn't derive parameters for received packet or timed out while waiting for remaining packet. Applicable NonTerminal: {applicable_nt} Received part: {complete_msg!r}. Exception: {str(parameter_parsing_exception)}"
            )
        else:
            nt_list = map(
                lambda x: x.format_as_spec(), forecast_non_terminals.get_non_terminals()
            )
            raise FandangoValueError(
                "Could not parse received message fragments into predicted NonTerminals. "
                + "Predicted NonTerminals: "
                + str(" | ".join(nt_list))
                + " Messages: "
                + str(io_instance.get_received_msgs())
            )

    max_parse_idx = -1
    best_parse_tree = None
    best_non_terminal = None
    for non_terminal, (parse_idx, parse_tree) in complete_parses.items():
        if max_parse_idx < parse_idx:
            max_parse_idx = parse_idx
            best_parse_tree = parse_tree
            best_non_terminal = non_terminal

    assert best_non_terminal is not None

    io_instance.clear_by_party(msg_sender, max_parse_idx)
    return forecast_non_terminals[best_non_terminal], best_parse_tree



import argparse
import difflib
import logging
import os
from typing import IO, Any
import zipfile

from fandango.constraints.soft import SoftValue
from fandango.constraints.constraint import Constraint
from fandango.errors import FandangoError, FandangoParseError
from fandango.language.grammar import ParsingMode
from fandango.language.grammar.grammar import Grammar
from fandango.language.parse import parse
from fandango.language.tree import DerivationTree
from fandango.logger import LOGGER, set_visualization


def report_syntax_error(
    filename: str, position: int, individual: str | bytes, *, binary: bool = False
) -> str:
    """
    Return position and error message in `individual`
    in user-friendly format.
    """
    if position >= len(individual):
        return f"{filename!r}: missing input at end of file"

    mismatch = individual[position]
    if binary:
        assert isinstance(mismatch, int)
        return f"{filename!r}, position {position:#06x} ({position}): mismatched input {mismatch.to_bytes()!r}"

    line = 1
    column = 1
    for i in range(position):
        if individual[i] == "\n":
            line += 1
            column = 1
        else:
            column += 1
    return f"{filename!r}, line {line}, column {column}: mismatched input {mismatch!r}"


def extract_initial_population(path: str) -> list[str]:
    try:
        initial_population = list()
        if path.strip().endswith(".zip"):
            with zipfile.ZipFile(path, "r") as zip:
                for file in zip.namelist():
                    data = zip.read(file).decode()
                    initial_population.append(data)
        else:
            for file in os.listdir(path):
                filename = os.path.join(path, file)
                with open(filename, "r") as fd:
                    individual = fd.read()
                initial_population.append(individual)
        return initial_population
    except FileNotFoundError as e:
        raise e


def _copy_setting(
    args: argparse.Namespace,
    settings: dict[str, Any],
    name: str,
    *,
    args_name: str | None = None,
) -> None:
    if args_name is None:
        args_name = name
    if hasattr(args, args_name) and getattr(args, args_name) is not None:
        settings[name] = getattr(args, args_name)
        LOGGER.debug(f"Settings: {name} is {settings[name]}")


def make_fandango_settings(
    args: argparse.Namespace, initial_settings: dict[str, Any] = {}
) -> dict[str, Any]:
    """Create keyword settings for Fandango() constructor"""
    LOGGER.debug(f"Pre-sanitized settings: {args}")
    settings = initial_settings.copy()
    _copy_setting(args, settings, "population_size")
    _copy_setting(args, settings, "mutation_rate")
    _copy_setting(args, settings, "crossover_rate")
    _copy_setting(args, settings, "elitism_rate")
    _copy_setting(args, settings, "destruction_rate")
    _copy_setting(args, settings, "warnings_are_errors")
    _copy_setting(args, settings, "best_effort")
    _copy_setting(args, settings, "random_seed")
    _copy_setting(args, settings, "max_repetition_rate")
    _copy_setting(args, settings, "max_repetitions")
    _copy_setting(args, settings, "max_nodes")
    _copy_setting(args, settings, "max_node_rate")

    if hasattr(args, "start_symbol") and args.start_symbol is not None:
        if args.start_symbol.startswith("<"):
            start_symbol = args.start_symbol
        else:
            start_symbol = f"<{args.start_symbol}>"
        settings["start_symbol"] = start_symbol

    if args.quiet and args.quiet == 1:
        LOGGER.setLevel(logging.WARNING)  # Default
    elif args.quiet and args.quiet > 1:
        LOGGER.setLevel(logging.ERROR)  # Even quieter
    elif args.verbose and args.verbose == 1:
        LOGGER.setLevel(logging.INFO)  # Give more info
    elif args.verbose and args.verbose > 1:
        LOGGER.setLevel(logging.DEBUG)  # Even more info

    if hasattr(args, "progress_bar") and args.progress_bar is not None:
        match args.progress_bar:
            case "on":
                set_visualization(True)
            case "off":
                set_visualization(False)
            case "auto":
                if args.infinite:
                    set_visualization(False)
                else:
                    set_visualization(None)

    if hasattr(args, "initial_population") and args.initial_population is not None:
        settings["initial_population"] = extract_initial_population(
            args.initial_population
        )
    return settings


def get_file_mode(
    args: argparse.Namespace,
    settings: dict[str, Any],
    *,
    grammar: Grammar | None = None,
    tree: DerivationTree | None = None,
) -> str:
    if (
        hasattr(args, "file_mode")
        and isinstance(args.file_mode, str)
        and args.file_mode != "auto"
    ):
        return args.file_mode

    if grammar is not None:
        start_symbol = settings.get("start_symbol", "<start>")
        if grammar.contains_bits(start=start_symbol) or grammar.contains_bytes(
            start=start_symbol
        ):
            return "binary"
        else:
            return "text"

    if tree is not None:
        if tree.should_be_serialized_to_bytes():
            return "binary"
        return "text"

    raise FandangoError("Cannot determine file mode")


def parse_contents_from_args(
    args: argparse.Namespace,
    given_grammars: list[Grammar] = [],
    check: bool = True,
) -> tuple[Grammar | None, list[Constraint | SoftValue]]:
    """Parse .fan content as given in args"""
    max_constraints = [f"maximizing {c}" for c in (args.maxconstraints or [])]
    min_constraints = [f"minimizing {c}" for c in (args.minconstraints or [])]
    constraints = (args.constraints or []) + max_constraints + min_constraints

    extra_defs = ""
    if "test_command" in args and args.test_command:
        arg_list = ", ".join(repr(arg) for arg in [args.test_command] + args.test_args)
        extra_defs += f"""
set_program_command([{arg_list}])
"""

    if "client" in args and args.client:
        # Act as client
        extra_defs += f"""
class Client(ConnectParty):
    def __init__(self):
        super().__init__(
            "{args.client}",
            ownership=Ownership.FANDANGO_PARTY,
            endpoint_type=EndpointType.CONNECT,
        )
        self.start()

class Server(ConnectParty):
    def __init__(self):
        super().__init__(
            "{args.client}",
            ownership=Ownership.EXTERNAL_PARTY,
            endpoint_type=EndpointType.OPEN,
        )
        self.start()
"""

    if "server" in args and args.server:
        # Act as server
        extra_defs += f"""
class Client(ConnectParty):
    def __init__(self):
        super().__init__(
            "{args.server}",
            ownership=Ownership.EXTERNAL_PARTY,
            endpoint_type=EndpointType.CONNECT,
        )
        self.start()

class Server(ConnectParty):
    def __init__(self):
        super().__init__(
            "{args.server}",
            ownership=Ownership.FANDANGO_PARTY,
            endpoint_type=EndpointType.OPEN,
        )
        self.start()
"""

    LOGGER.debug("Extra definitions:" + extra_defs)
    args.fan_files += [extra_defs]

    return parse(
        args.fan_files,
        constraints,
        given_grammars=given_grammars,
        includes=args.includes,
        use_cache=args.use_cache,
        use_stdlib=args.use_stdlib,
        start_symbol=args.start_symbol,
        parties=args.parties,
        check=check,
    )


def parse_constraints_from_args(
    args: argparse.Namespace,
    given_grammars: list[Grammar] = [],
    check: bool = True,
) -> tuple[Grammar | None, list[Constraint | SoftValue]]:
    """Parse .fan constraints as given in args"""
    max_constraints = [f"maximizing {c}" for c in (args.maxconstraints or [])]
    min_constraints = [f"minimizing {c}" for c in (args.minconstraints or [])]
    constraints = (args.constraints or []) + max_constraints + min_constraints
    return parse(
        [],
        constraints,
        given_grammars=given_grammars,
        includes=args.includes,
        use_cache=args.use_cache,
        use_stdlib=args.use_stdlib,
        start_symbol=args.start_symbol,
        parties=args.parties,
        check=check,
    )


def validate(
    original: str | bytes | DerivationTree,
    parsed: DerivationTree,
    *,
    filename: str = "<file>",
) -> None:
    if (
        (isinstance(original, DerivationTree) and original.value() != parsed.value())
        or (isinstance(original, bytes) and original != parsed.to_bytes())
        or (isinstance(original, str) and original != parsed.to_string())
    ):
        exc = FandangoError(f"{filename!r}: parsed tree does not match original")
        if getattr(Exception, "add_note", None):
            # Python 3.11+ has add_note() method
            if isinstance(original, DerivationTree) and isinstance(
                parsed, DerivationTree
            ):
                original_grammar = original.to_grammar()
                parsed_grammar = parsed.to_grammar()
                diff = difflib.context_diff(
                    original_grammar.split("\n"),
                    parsed_grammar.split("\n"),
                    fromfile="original",
                    tofile="parsed",
                )
                out = "\n".join(line for line in diff)
                exc.add_note(out)
        raise exc


def parse_file(
    fd: IO[Any],
    args: argparse.Namespace,
    grammar: Grammar,
    constraints: list[Constraint | SoftValue],
    settings: dict[str, Any],
) -> DerivationTree:
    """
    Parse a single file `fd` according to `args`, `grammar`, `constraints`, and `settings`, and return the parse tree.
    """
    LOGGER.info(f"Parsing {fd.name!r}")
    individual = fd.read()
    start_symbol = settings.get("start_symbol", "<start>")

    allow_incomplete = hasattr(args, "prefix") and args.prefix
    parsing_mode = ParsingMode.COMPLETE
    if allow_incomplete:
        parsing_mode = ParsingMode.INCOMPLETE
    tree_gen = grammar.parse_forest(individual, start=start_symbol, mode=parsing_mode)

    alternative_counter = 1
    passing_tree = None
    last_tree = None
    while tree := next(tree_gen, None):
        LOGGER.debug(f"Checking parse alternative #{alternative_counter}")
        last_tree = tree
        grammar.populate_sources(last_tree)

        passed = True
        for constraint in constraints:
            fitness = constraint.fitness(tree).fitness()
            LOGGER.debug(f"Fitness: {fitness}")
            if fitness == 0:
                passed = False
                break

        if passed:
            passing_tree = tree
            break

        # Try next parsing alternative
        alternative_counter += 1

    if passing_tree:
        # Found an alternative that satisfies all constraints

        # Validate tree
        if args.validate:
            validate(individual, passing_tree, filename=fd.name)

        return passing_tree

    # Tried all alternatives
    if last_tree is None:
        error_pos = grammar.max_position() + 1
        raise FandangoParseError(
            report_syntax_error(fd.name, error_pos, individual, binary=("b" in fd.mode))
        )

    # Report error for the last tree
    for constraint in constraints:
        fitness = constraint.fitness(last_tree).fitness()
        if fitness == 0:
            raise FandangoError(f"{fd.name!r}: constraint {constraint} not satisfied")

    raise FandangoError("This should not happen")


def exec_single(
    code: str, _globals: dict[str, Any] = {}, _locals: dict[str, Any] = {}
) -> None:
    """Execute CODE in 'single' mode, printing out results if any"""
    block = compile(code, "<input>", mode="single")
    exec(block, _globals, _locals)



from fandango.io.packetforecaster import PacketForecaster
from fandango.language.grammar import ParsingMode
from fandango.language.symbols import NonTerminal
from fandango.language.tree import DerivationTree
from tests.utils import RESOURCES_ROOT
from fandango.api import Fandango


def get_grammar():
    with open(RESOURCES_ROOT / "forecaster.fan") as f:
        spec = f.read()
    fandango = Fandango(spec, use_stdlib=False, use_cache=False)
    return fandango.grammar


def assert_prediction(prediction, expected: dict[str, list[str]]):
    assert len(prediction.get_msg_parties()) == len(expected)
    for role, nonterminals in expected.items():
        assert role in prediction
        assert len(nonterminals) == len(
            prediction.parties_to_packets[role].nt_to_packet
        )
        for nt in nonterminals:
            assert NonTerminal(nt) in prediction.parties_to_packets[role].nt_to_packet


def test_forecast_1():
    grammar = get_grammar()
    forecaster = PacketForecaster(grammar)
    tree = DerivationTree(NonTerminal("<start>"))
    prediction: PacketForecaster.ForecastingResult = forecaster.predict(tree)
    expected = {"StdOut": ["<d>", "<e>", "<c>"]}
    assert_prediction(prediction, expected)


def test_forecast_2():
    grammar = get_grammar()
    forecaster = PacketForecaster(grammar)
    tree = grammar.parse("d", mode=ParsingMode.INCOMPLETE)
    assert tree is not None
    prediction: PacketForecaster.ForecastingResult = forecaster.predict(tree)
    expected = {"StdOut": ["<e>", "<c>"]}
    assert_prediction(prediction, expected)


def test_forecast_3():
    grammar = get_grammar()
    forecaster = PacketForecaster(grammar)
    tree = grammar.parse("de", mode=ParsingMode.INCOMPLETE)
    assert tree is not None
    prediction: PacketForecaster.ForecastingResult = forecaster.predict(tree)
    expected = {"StdOut": ["<e>", "<c>"]}
    assert_prediction(prediction, expected)


def test_forecast_4():
    grammar = get_grammar()
    forecaster = PacketForecaster(grammar)
    tree = grammar.parse("dec", mode=ParsingMode.INCOMPLETE)
    assert tree is not None
    prediction: PacketForecaster.ForecastingResult = forecaster.predict(tree)
    expected = {"StdOut": ["<c>", "<g>", "<i>"]}
    assert_prediction(prediction, expected)


def test_forecast_5():
    grammar = get_grammar()
    forecaster = PacketForecaster(grammar)
    tree = grammar.parse("dc", mode=ParsingMode.INCOMPLETE)
    assert tree is not None
    prediction: PacketForecaster.ForecastingResult = forecaster.predict(tree)
    expected = {"StdOut": ["<c>", "<g>", "<i>"]}
    assert_prediction(prediction, expected)


def test_forecast_6():
    grammar = get_grammar()
    forecaster = PacketForecaster(grammar)
    tree = grammar.parse("dcc", mode=ParsingMode.INCOMPLETE)
    assert tree is not None
    prediction: PacketForecaster.ForecastingResult = forecaster.predict(tree)
    expected = {"StdOut": ["<g>", "<i>"]}
    assert_prediction(prediction, expected)


def test_forecast_7():
    grammar = get_grammar()
    forecaster = PacketForecaster(grammar)
    tree = grammar.parse("dccg", mode=ParsingMode.INCOMPLETE)
    assert tree is not None
    prediction: PacketForecaster.ForecastingResult = forecaster.predict(tree)
    expected: dict[str, list[str]] = {}
    assert_prediction(prediction, expected)


def test_forecast_8():
    grammar = get_grammar()
    forecaster = PacketForecaster(grammar)
    tree = grammar.parse("dcci", mode=ParsingMode.INCOMPLETE)
    assert tree is not None
    prediction: PacketForecaster.ForecastingResult = forecaster.predict(tree)
    expected: dict[str, list[str]] = {}
    assert_prediction(prediction, expected)



from fandango.language.symbols.symbol import Symbol, SymbolType
from fandango.language.symbols.non_terminal import NonTerminal
from fandango.language.symbols.terminal import Terminal
from fandango.language.symbols.slice import Slice

__all__ = ["Symbol", "SymbolType", "NonTerminal", "Terminal", "Slice"]



from io import UnsupportedOperation
import re
from types import NoneType

import regex

from fandango.errors import FandangoValueError
from fandango.language.symbols import Symbol, SymbolType
from fandango.language.tree_value import TreeValue, TreeValueType


class Terminal(Symbol):
    def __init__(self, symbol: str | bytes | int | TreeValue) -> None:
        super().__init__(symbol, SymbolType.TERMINAL)

    def count_bytes(self) -> int:
        return self._value.count_bytes()

    @staticmethod
    def string_prefix(symbol: str) -> str:
        """Return the first letters ('f', 'b', 'r', ...) of a string literal"""
        match = re.match(r"([a-zA-Z]+)", symbol)
        return match.group(0) if match else ""

    @staticmethod
    def clean(symbol: str) -> str | bytes | int:
        # LOGGER.debug(f"Cleaning {symbol!r}")
        if symbol.startswith("f'") or symbol.startswith('f"'):
            # Cannot evaluate f-strings
            raise UnsupportedOperation("f-strings are currently not supported")

        return eval(symbol)  # also handles bits "0" and "1"

    @staticmethod
    def from_symbol(symbol: str) -> "Terminal":
        t = Terminal(Terminal.clean(symbol))
        t._is_regex = "r" in Terminal.string_prefix(symbol)
        return t

    @staticmethod
    def from_number(number: str) -> "Terminal":
        return Terminal(Terminal.clean(number))

    def check(self, word: str | bytes | int, incomplete=False) -> tuple[bool, int]:
        """Return (True, # characters matched by `word`), or (False, 0)"""

        if self._value.is_type(TreeValueType.TRAILING_BITS_ONLY) or isinstance(
            word, int
        ):
            return self.check_all(word), 1
        check_word: str | bytes = word

        symbol: str | bytes
        if self._value.is_type(TreeValueType.BYTES) and isinstance(check_word, bytes):
            symbol = self._value.to_bytes()
        else:
            symbol = self._value.to_string()
            check_word = (
                check_word
                if isinstance(check_word, str)
                else check_word.decode("latin-1")
            )

        if self.is_regex:
            if not incomplete:
                match = re.match(symbol, check_word)  # type: ignore [arg-type] # re actually does accept bytes
                if match:
                    # LOGGER.debug(f"It's a match: {match.group(0)!r}")
                    return True, len(match.group(0))
            else:
                compiled = regex.compile(symbol)
                match = compiled.match(check_word, partial=True)
                if match is not None and (
                    match.partial or match.end() == len(check_word)
                ):
                    return True, len(match.group(0))
                match = compiled.fullmatch(check_word, partial=True)
                if match is not None and (
                    match.partial or match.end() == len(check_word)
                ):
                    return True, len(match.group(0))
                return False, 0

        else:
            if incomplete:
                prefix = check_word
                full_word = symbol
            else:
                prefix = symbol
                full_word = check_word
            if isinstance(full_word, str):
                assert isinstance(prefix, str)
                if full_word.startswith(prefix):
                    return True, len(prefix)
            else:
                assert isinstance(full_word, bytes)
                assert isinstance(prefix, bytes)
                if full_word.startswith(prefix):
                    return True, len(prefix)

        # LOGGER.debug(f"No match")
        return False, 0

    def check_all(self, word: str | bytes | int) -> bool:
        if isinstance(word, str):
            return self._value.to_string() == word
        elif isinstance(word, bytes):
            return self._value.to_bytes() == word
        elif isinstance(word, int):
            return int(self._value) == word
        else:
            raise FandangoValueError(f"Invalid word type: {type(word)}")

    def format_as_spec(self) -> str:
        if self.is_regex:
            if self.is_type(TreeValueType.BYTES):
                symbol = repr(self._value)
                symbol = symbol.replace(r"\\", "\\")
                return "r" + symbol
            elif self.is_type(TreeValueType.TRAILING_BITS_ONLY):
                return "r'" + str(self._value) + "'"

            if "'" not in str(self._value):
                return "r'" + str(self._value) + "'"
            if '"' not in str(self._value):
                return 'r"' + str(self._value) + '"'

            # Mixed quotes: encode single quotes
            symbol = str(self._value).replace("'", r"\x27")
            return "r'" + str(symbol) + "'"

        # Not a regex
        return repr(self._value)

    def __hash__(self) -> int:
        return hash((self._value, self._type))

    def __repr__(self) -> str:
        return "Terminal(" + self.format_as_spec() + ")"

    def __len__(self) -> int:
        return self.count_bytes()



import glob
from io import StringIO
import os
import re
import sys

from fandango.cli.commands import COMMANDS
from fandango.cli.parser import get_parser
from fandango.logger import LOGGER


def complete(text: str) -> list[str]:
    """Return possible completions for TEXT"""
    LOGGER.debug("Completing " + repr(text))

    if not text:
        # No text entered, all commands possible
        completions = [s for s in COMMANDS.keys()]
        LOGGER.debug("Completions: " + repr(completions))
        return completions

    completions = []
    for s in COMMANDS.keys():
        if s.startswith(text):
            completions.append(s + " ")
    if completions:
        # Beginning of command entered
        LOGGER.debug("Completions: " + repr(completions))
        return completions

    # Complete command
    words = text.split()
    cmd = words[0]
    shell = cmd.startswith("!") or cmd.startswith("/")

    if not shell and cmd not in COMMANDS.keys():
        # Unknown command
        return []

    if len(words) == 1 or text.endswith(" "):
        last_arg = ""
    else:
        last_arg = words[-1]

    # print(f"last_arg = {last_arg}")
    completions = []

    if not shell:
        cmd_options = get_options(cmd)
        for option in cmd_options:
            if not last_arg or option.startswith(last_arg):
                completions.append(option + " ")

    if shell or len(words) >= 2:
        # Argument for an option
        filenames = get_filenames(prefix=last_arg, fan_only=not shell)
        for filename in filenames:
            if filename.endswith(os.sep):
                completions.append(filename)
            else:
                completions.append(filename + " ")

    LOGGER.debug("Completions: " + repr(completions))
    return completions


def get_filenames(prefix: str = "", fan_only: bool = True) -> list[str]:
    """Return all files that match PREFIX"""
    filenames = []
    all_filenames = glob.glob(prefix + "*")
    for filename in all_filenames:
        if os.path.isdir(filename):
            filenames.append(filename + os.sep)
        elif (
            not fan_only
            or filename.lower().endswith(".fan")
            or filename.lower().endswith(".py")
        ):
            filenames.append(filename)

    return filenames


def get_help(cmd: str) -> str:
    """Return the help text for CMD"""
    parser = get_parser(in_command_line=False)
    old_stdout = sys.stdout
    sys.stdout = mystdout = StringIO()

    try:
        parser.parse_args([cmd] + ["--help"])
    except SystemExit:
        pass

    sys.stdout = old_stdout
    return mystdout.getvalue()


def get_options(cmd: str) -> list[str]:
    """Return all --options for CMD"""
    if cmd == "help":
        return list(COMMANDS.keys())

    help = get_help(cmd)
    options = []
    for option in re.findall(r"--?[a-zA-Z0-9_-]*", help):
        if option not in options:
            options.append(option)
    return options



from fandango.language.grammar.node_visitors.node_visitor import NodeVisitor
from fandango.language.grammar.nodes.non_terminal import NonTerminalNode
from fandango.language.grammar.nodes.terminal import TerminalNode


class SymbolFinder(NodeVisitor):
    def __init__(self):
        self.terminalNodes = []
        self.nonTerminalNodes = []

    def visitNonTerminalNode(self, node: NonTerminalNode):
        self.nonTerminalNodes.append(node)

    def visitTerminalNode(self, node: TerminalNode):
        self.terminalNodes.append(node)



#!/usr/bin/env pytest

import ast
import os

import pytest
from antlr4 import InputStream, CommonTokenStream, BailErrorStrategy
from antlr4.error.Errors import ParseCancellationException

from fandango.constraints import predicates
from fandango.constraints.comparison import ComparisonConstraint
from fandango.constraints.failing_tree import Comparison
from fandango.language.convert import (
    FandangoSplitter,
    GrammarProcessor,
    SearchProcessor,
    PythonProcessor,
)
from fandango.language.grammar.grammar import Grammar
from fandango.language.grammar.nodes.alternative import Alternative
from fandango.language.parse import parse
from fandango.language.parser.FandangoLexer import FandangoLexer
from fandango.language.parser.FandangoParser import FandangoParser
from fandango.language.search import RuleSearch
from fandango.language.symbols import NonTerminal
from .utils import RESOURCES_ROOT

FUZZINGBOOK_GRAMMAR = {
    "<start>": ["<number>"],
    "<number>": ["<non_zero><digits>", "0"],
    "<digits>": ["", "<digit><digits>"],
    "<non_zero>": ["1", "2", "3", "4", "5", "6", "7", "8", "9"],
    "<digit>": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
}


def get_tree(example, start="fandango"):
    lexer = FandangoLexer(InputStream(example))
    token = CommonTokenStream(lexer)
    parser = FandangoParser(token)
    parser._errHandler = BailErrorStrategy()
    return getattr(parser, start)()


def test_indents():
    tree = get_tree(
        r"""
<a> ::= \
    "a" \
    | "a" <a>
"""
    )
    splitter = FandangoSplitter()
    splitter.visit(tree)
    processor = GrammarProcessor(grammar_settings=splitter.grammar_settings)
    grammar = processor.get_grammar(splitter.productions)
    assert len(grammar.rules) == 1
    rule = list(grammar.rules.values())[0]
    assert isinstance(rule, Alternative)
    assert len(rule.alternatives) == 2


def test_alt_indents():
    tree = get_tree(
        r"""
<a> ::= (
    "a"
    | "a" <a>
    )
"""
    )
    splitter = FandangoSplitter()
    splitter.visit(tree)
    processor = GrammarProcessor(grammar_settings=splitter.grammar_settings)
    grammar = processor.get_grammar(splitter.productions)
    assert len(grammar.rules) == 1
    rule = list(grammar.rules.values())[0]
    assert isinstance(rule, Alternative)
    assert len(rule.alternatives) == 2


@pytest.mark.parametrize(
    "expression",
    [
        "x",
        "1",
        "x and y",
        "x or y",
        "x + y",
        "x - y",
        "x * y",
        "x / y",
        "x // y",
        "x ** y",
        "x @ y",
        "x << y",
        "x >> y",
        "x | y",
        "x ^ y",
        "~ x",
        "not x",
        "+ x",
        "- x",
        "x if y else z",
        "{x: y, v: w}",
        "{x, y, z}",
        "[x, y, z]",
        "(x, y, z)",
        "{x: y for x in z if x}",
        "{x for x in z if x}",
        "[x for x in z if x]",
        "(x for x in z if x)",
        "await x",
        "yield x",
        "yield from x",
        "x < y",
        "x <= y",
        "x > y",
        "x >= y",
        "x == y",
        # "x <> y",
        "x != y",
        "x is y",
        "x is not y",
        "x in y",
        "x not in y",
        "x()",
        "x(y, z, *a, v=w, **k)",
        "x.y.z",
        "x[y,a:b:c,::]",
        "*x",
    ],
)
def test_conversion_without_replace(expression):
    try:
        tree = ast.parse(expression, mode="eval")
    except SyntaxError:
        with pytest.raises(ParseCancellationException):
            get_tree(
                expression,
                start="expression",
            )
    else:
        fandango_tree: FandangoParser.ExpressionContext = get_tree(
            expression, start="expression"
        )
        processor = SearchProcessor(Grammar.dummy())
        fandango_tree, searches, search_map = processor.visit(fandango_tree)
        assert 0 == len(searches)
        assert 0 == len(search_map)
        assert not isinstance(fandango_tree, list)
        assert ast.unparse(tree) == ast.unparse(fandango_tree)


@pytest.mark.parametrize(
    "expression,value,expected",
    [
        ("<x>", True, True),
        ("<x> and True", True, True),
        ("<x> or False", True, True),
        ("<x> + 3", 2, 5),
        ("<x> - 3", 2, -1),
        ("<x> * 3", 2, 6),
        ("<x> / 2", 2, 1),
        ("<x> // 2", 3, 1),
        ("<x> ** 3", 2, 8),
        ("<x> << 2", 2, 8),
        ("<x> >> 2", 8, 2),
        ("<x> | 2", 1, 3),
        ("<x> ^ 5", 4, 1),
        ("~ <x>", 1, -2),
        ("not <x>", True, False),
        ("+ <x>", 2, 2),
        ("- <x>", 2, -2),
        ("2 if <x> else 1", True, 2),
        ("{1: <x>, 3: 4}", 2, {1: 2, 3: 4}),
        ("{1, <x>, 3}", 2, {1, 2, 3}),
        ("[1, <x>, 3]", 2, [1, 2, 3]),
        ("(1, <x>, 3)", 2, (1, 2, 3)),
        ("{x: x for x in <x> if x % 2 == 1}", [1, 2, 3], {1: 1, 3: 3}),
        ("{x for x in <x> if x % 2 == 1}", [1, 2, 3], {1, 3}),
        ("[x for x in <x> if x % 2 == 1]", [1, 2, 3], [1, 3]),
        ("tuple(x for x in <x> if x % 2 == 1)", [1, 2, 3], (1, 3)),
        ("<x> < 3", 2, True),
        ("<x> <= 3", 2, True),
        ("<x> > 3", 2, False),
        ("<x> >= 3", 2, False),
        ("<x> == 3", 2, False),
        ("<x> != 3", 2, True),
        ("<x> is None", 2, False),
        ("<x> is not None", 2, True),
        ("<x> in [1, 2, 3]", 2, True),
        ("<x> not in [1, 2, 3]", 2, False),
        ("int(<x>)", "2", 2),
        ("<x>.lower()", "AbC", "abc"),
        # ("<x>[0, 2::2]", [1, 2, 3, 4], [1, 2, 4]),
    ],
)
def test_conversion_with_replacement(expression, value, expected):
    fandango_tree: FandangoParser.ExpressionContext = get_tree(
        expression, start="expression"
    )
    processor = SearchProcessor(Grammar.dummy())
    fandango_tree, searches, search_map = processor.visit(fandango_tree)
    assert 1 == len(search_map)
    assert 1 == len(searches)
    placeholder = list(search_map.keys())[0]
    assert expected == eval(ast.unparse(fandango_tree), {}, {placeholder: value})


@pytest.mark.parametrize(
    "stmt,value,is_global",
    [
        ("x = y = 1", 1, False),
        ("x: int = 1", 1, False),
        ("x = 0\nx += 1", 1, False),
        ("x = [0]\nx[0] += 1\nx = x[0]", 1, False),
        ("import os\nx = os.path.join('a', 'b')", os.path.join("a", "b"), False),
        (
            "try:\n    raise ValueError()\nexcept:\n    x = 1\nelse:\n    x = 1",
            1,
            False,
        ),
        ("a = [0, 1]\ndel a[0]\nx = a[0]", 1, False),
        ("pass\nx = 1", 1, False),
        ("def f(): yield 1\nfor x in f(): pass", 1, False),
        ("assert True\nx = 1", 1, False),
        ("for x in range(4):\n    if x != 1: continue\n    else: break", 1, False),
        ("global x\nx = 1", 1, True),
        (
            "def f():\n"
            "    y = 0\n"
            "    def g():\n"
            "        nonlocal y\n"
            "        y = 1\n"
            "        return y\n"
            "    return y + g()\n"
            "x = f()",
            1,
            False,
        ),
        ("if False: x = 0\nelif False: x = 2\nelse: x = 1", 1, False),
        ("class A: x = 1\nx = A.x", 1, False),
        (
            "class A:\n"
            "    def __enter__(self): return 1\n"
            "    def __exit__(self, exc_type, exc_val, exc_tb): pass\n"
            "with A() as x: pass",
            1,
            False,
        ),
        ("x = -2\nwhile x != 1: x += 1", 1, False),
        (
            "def f(y, *a, z: int, b: int=2, **c): return y + z - b\nx = f(4, z=-1)",
            1,
            False,
        ),
    ],
)
def test_conversion_statement(stmt, value, is_global):
    is_global = True  # As of now, all Python defs are parsed as global defs

    fandango_tree: FandangoParser.ExpressionContext = get_tree(stmt)
    splitter = FandangoSplitter()
    splitter.visit(fandango_tree)
    code = splitter.python_code
    processor = PythonProcessor()
    fandango_tree = processor.get_code(code)
    tree = ast.parse(stmt)
    assert ast.unparse(fandango_tree) == ast.unparse(tree)
    global_vars = predicates.__dict__.copy()
    local_vars = None  # Must be None to ensure top-level imports
    exec(ast.unparse(fandango_tree), global_vars, local_vars)
    if is_global:
        assert "x" in global_vars
        assert local_vars is None or "x" not in local_vars
        assert value == global_vars["x"]
    else:
        assert "x" not in global_vars
        assert local_vars is not None and "x" in local_vars
        assert value == local_vars["x"]


def test_parsing():
    with open(RESOURCES_ROOT / "fandango.fan", "r") as file:
        grammar, constraints = parse(file, use_stdlib=False, use_cache=False)
    assert isinstance(grammar, Grammar)
    assert len(grammar.rules) == 4
    assert "<start>" in grammar
    assert "<number>" in grammar
    assert "<non_zero>" in grammar
    assert "<digit>" in grammar
    assert len(constraints) == 1
    constraint = constraints[0]
    assert isinstance(constraint, ComparisonConstraint)
    assert constraint.right == "0"
    assert constraint.operator == Comparison.EQUAL
    assert len(constraint.searches) == 1
    placeholder = list(constraint.searches.keys())[0]
    assert constraint.left == f"f({placeholder}) % 2"
    assert "f" in constraint.global_variables
    assert eval("f('1')", constraint.global_variables, constraint.local_variables) == 1
    search = constraint.searches[placeholder]
    assert isinstance(search, RuleSearch)
    assert NonTerminal("<number>") == search.symbol



#!/usr/bin/env python

import argparse
import re
import sys
from typing import Any

from lxml import etree  # type: ignore[attr-defined] # types not available

from fandango.converters.FandangoConverter import FandangoConverter


def fan(name):
    """Convert a name to a Fandango identifier."""
    return re.sub(r"[^a-zA-Z0-9_]", "_", name)


# See https://lxml.de/validation.html#dtd-1 for information on DTD structure


class DTDFandangoConverter(FandangoConverter):
    """Convert a DTD schema to a Fandango grammar."""

    def __init__(self, filename: str):
        super().__init__(filename)

        self.dtd = etree.DTD(filename)
        self.entities = {}

        for entity in self.dtd.iterentities():
            # print(f"name = {entity.name}, orig = {entity.orig}, content = {entity.content}")
            self.entities[entity.name] = entity.content

        self.attribute_types: dict[str, str] = {}

    def header(self):
        s = f"""# Automatically generated from {self.filename!r}.
#
<ws> ::= <whitespace>+ := ' '  # whitespace sequence
<q> ::= ('"' | "'")? := '"'    # optional quote
"""
        return s

    def to_fan(self) -> str:
        self.attribute_types = {}

        s = self.header()
        for element in self.dtd.iterelements():
            s += self.convert_element(element)

        types = list(self.attribute_types)
        types.sort()
        if types:
            s += "\n\n# Attribute types, to be further refined"
        for tp in types:
            s += f"\n<{tp}> ::= {self.attribute_types[tp]}"

        return s

    def convert_element(self, element) -> str:
        attrs, values, required_attributes = self.convert_attributes(element)

        s = f"\n\n# {element.name} ({element.type})\n"
        s += f"<{fan(element.name)}> ::= '<{element.name}' (<ws> <{fan(element.name)}_attribute>)* ('/>' | '>' "
        s += self.convert_content(element.content)
        s += f" '</{fan(element.name)}>')\n"
        s += attrs

        if required_attributes:
            s += "\nwhere ("
            s += "\n   and ".join(
                f"{fan(attribute.name) + '='!r} in <{fan(element.name)}>.descendant_values()"
                for attribute in required_attributes
            )
            s += ")  # required"
        if values:
            s += f"\n\n# {element.name} attribute types"
        for value in values:
            s += f"\n<{fan(element.name)}_{value}> ::= <{value}>"
        return s

    def convert_content(self, content):
        if content is None:
            return ""

        s = ""
        match content.type:
            case "pcdata":
                s += "<pcdata>"
            case "element":
                s += f"<{fan(content.name)}>"
            case "seq":
                s += (
                    self.convert_content(content.left)
                    + " "
                    + self.convert_content(content.right)
                )
            case "or":
                s += (
                    self.convert_content(content.left)
                    + " | "
                    + self.convert_content(content.right)
                )
            case _:
                raise ValueError(f"Unknown content type {content.type!r}")

        match content.occur:
            case "once":
                pass
            case "opt":
                s = f"({s})?"
            case "mult":
                s = f"({s})*"
            case "plus":
                s = f"({s})+"
            case _:
                raise ValueError(f"Unknown occurrence {content.occur!r}")

        return s

    def convert_attributes(self, element) -> tuple[str, list[str], list[Any]]:
        s = f"<{fan(element.name)}_attribute> ::= "

        attrs = []
        values = []
        required_attributes = []
        for attribute in element.iterattributes():
            attr, value, required = self.convert_attribute(attribute)
            attrs.append(attr)
            if value:
                values.append(value)
            if required:
                required_attributes.append(attribute)

        s += "\n    | ".join(attrs)
        return s, values, required_attributes

    def convert_attribute(self, attribute) -> tuple[str, str | None, bool]:
        value = None
        s = f"'{fan(attribute.name)}='"
        if attribute.default == "fixed":
            s += f" <q> {attribute.default_value!r} <q>"
            return s, None, False

        match attribute.type:
            case "enumeration":
                values = (
                    " <q> ("
                    + " | ".join(f"{value!r}" for value in attribute.itervalues())
                    + ") <q>"
                )
                s += values
            case _:
                value = fan(attribute.name + "_value")
                self.attribute_types[value] = f"<{attribute.type}>"
                s += f" <{fan(attribute.elemname)}_{value}> "

        required = attribute.default == "required"
        s += f"  # {attribute.default}"
        if attribute.default_value:
            s += f"; default {attribute.default_value!r}"

        return s, value, required


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert a [DTD] schema to a Fandango specification"
    )

    parser.add_argument(dest="files", action="append", type=str, help="schema file")

    args = parser.parse_args(sys.argv[1:])

    for filename in args.files:
        converter = DTDFandangoConverter(filename)
        print(converter.to_fan())



# From https://github.com/antlr/grammars-v4/blob/master/antlr/antlr4/Python3/LexerAdaptor.py

# --------------------------------------------------------------------------------
# [The "BSD licence"]
# Copyright (c) 2005-2007 Terence Parr
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# --------------------------------------------------------------------------------

import sys
from typing import TextIO

from antlr4 import *

# NB!!!!!!!!!:
# Python3 is a terrible language. You cannot import ANTLRv4Lexer
# because Python3 cannot handle circular imports. So, we hardwire a
# bunch of constants directly in this damn code. If someone can
# figure out how to pound Python into submision and just work
# please fix.
# Cannot import lexer so use parser instead.
from fandango.converters.antlr.ANTLRv4Parser import ANTLRv4Parser


class LexerAdaptor(Lexer):
    # Constants (mirroring the Java static final ints)
    PREQUEL_CONSTRUCT = -10
    OPTIONS_CONSTRUCT = -11

    def __init__(self, input: InputStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        # Track whether we are in a parser rule, lexer rule, or outside any rule.
        # Token.INVALID_TYPE indicates we are outside any rule context.
        self._currentRuleType = Token.INVALID_TYPE

    def getCurrentRuleType(self):
        return self._currentRuleType

    def setCurrentRuleType(self, ruleType):
        self._currentRuleType = ruleType

    def handleBeginArgument(self):
        """
        Decide how to handle a '[' token. If inside a lexer rule, it may be a character set;
        otherwise, it's a normal argument list and we switch to Argument mode.
        """
        if self.inLexerRule():
            self.pushMode(2)
            # NB!!!!!!!!!: hardwire ANTLRv4Parser.LexerCharSet)
            self.more()
        else:
            self.pushMode(1)
            # NB!!!!!!!!!: hardwire ANTLRv4Lexer.Argument)

    def handleEndArgument(self):
        """
        Upon finding a matching ']', pop out of Argument mode. If we still have a mode stack
        (i.e., nested contexts), set the type to ARGUMENT_CONTENT.
        """
        self.popMode()
        if len(self._modeStack) > 0:
            self.type = ANTLRv4Parser.ARGUMENT_CONTENT

    def emit(self):
        """
        Custom emit logic to update _currentRuleType based on tokens like OPTIONS, TOKENS,
        CHANNELS, RBRACE, AT, SEMI, ID, etc.
        """
        if (
            self._type == ANTLRv4Parser.OPTIONS
            or self._type == ANTLRv4Parser.TOKENS
            or self._type == ANTLRv4Parser.CHANNELS
        ) and self.getCurrentRuleType() == Token.INVALID_TYPE:
            # Enter prequel construct block, which ends when we see a '}'
            self.setCurrentRuleType(self.PREQUEL_CONSTRUCT)

        elif (
            self._type == ANTLRv4Parser.OPTIONS
            and self.getCurrentRuleType() == ANTLRv4Parser.TOKEN_REF
        ):
            # We encountered OPTIONS inside a lexer rule, treat it specially
            self.setCurrentRuleType(self.OPTIONS_CONSTRUCT)

        elif (
            self._type == ANTLRv4Parser.RBRACE
            and self.getCurrentRuleType() == self.PREQUEL_CONSTRUCT
        ):
            # Exiting a prequel construct block
            self.setCurrentRuleType(Token.INVALID_TYPE)

        elif (
            self._type == ANTLRv4Parser.RBRACE
            and self.getCurrentRuleType() == self.OPTIONS_CONSTRUCT
        ):
            # Exiting an options block back into a lexer rule
            self.setCurrentRuleType(ANTLRv4Parser.TOKEN_REF)

        elif (
            self._type == ANTLRv4Parser.AT
            and self.getCurrentRuleType() == Token.INVALID_TYPE
        ):
            # Entering an action block
            self.setCurrentRuleType(ANTLRv4Parser.AT)

        elif (
            self._type == ANTLRv4Parser.SEMI
            and self.getCurrentRuleType() == self.OPTIONS_CONSTRUCT
        ):
            # A semicolon inside an options {...} block, do nothing special
            pass

        elif (
            self._type == ANTLRv4Parser.ACTION
            and self.getCurrentRuleType() == ANTLRv4Parser.AT
        ):
            # Exiting an action block
            self.setCurrentRuleType(Token.INVALID_TYPE)

        elif self._type == ANTLRv4Parser.ID:
            # Distinguish between TOKEN_REF (uppercase ID) vs. RULE_REF (lowercase ID)
            # If the input is only a single char, getText(...) should still return that char
            # You may need to import Interval or adapt how text is retrieved for your runtime
            from antlr4 import InputStream

            firstChar = self._input.getText(
                self._tokenStartCharIndex, self._tokenStartCharIndex
            )
            if firstChar and firstChar[0].isupper():
                self._type = ANTLRv4Parser.TOKEN_REF
            else:
                self._type = ANTLRv4Parser.RULE_REF

            # If we were outside a rule, now we're inside a rule of the indicated type
            if self.getCurrentRuleType() == Token.INVALID_TYPE:
                self.setCurrentRuleType(self._type)

        elif self._type == ANTLRv4Parser.SEMI:
            # The ';' token indicates the end of a rule definition
            self.setCurrentRuleType(Token.INVALID_TYPE)

        return super().emit()

    def inLexerRule(self):
        """
        True if we are currently inside a lexer rule (TOKEN_REF).
        """
        return self.getCurrentRuleType() == ANTLRv4Parser.TOKEN_REF

    def inParserRule(self):
        """
        True if we are currently inside a parser rule (RULE_REF).
        """
        return self.getCurrentRuleType() == ANTLRv4Parser.RULE_REF

    def reset(self):
        """
        Override reset to also reset the current rule type to INVALID_TYPE.
        """
        self.setCurrentRuleType(Token.INVALID_TYPE)
        super().reset()



#!/usr/bin/env pytest

from collections import Counter
import pytest

from fandango.evolution import havoc

ITERS = 1000
INPUT_SIZE = 100


def adjacent_diff(input_: bytearray) -> list[int]:
    return [input_[i] - input_[i - 1] for i in range(1, len(input_))]


def test_wrapping_add():
    from fandango.evolution.havoc import _wrapping_add

    assert _wrapping_add(1, 1) == 2
    assert _wrapping_add(256 - 1, 1) == 0
    assert _wrapping_add(256**2 - 1, 1, byte_size=2) == 0
    assert _wrapping_add(256**4 - 1, 1, byte_size=4) == 0


def test_wrapping_sub():
    from fandango.evolution.havoc import _wrapping_sub

    assert _wrapping_sub(1, 1) == 0
    assert _wrapping_sub(0, 1) == 256 - 1
    assert _wrapping_sub(0, 1, byte_size=2) == 256**2 - 1
    assert _wrapping_sub(0, 1, byte_size=4) == 256**4 - 1


def test_get_random_index():
    from fandango.evolution.havoc import _get_random_index

    input = bytearray([0] * INPUT_SIZE)
    for _ in range(ITERS):
        index = _get_random_index(input)
        assert index < INPUT_SIZE
        assert index >= 0
        assert input[index] == 0


def test_get_random_range_of_length():
    from fandango.evolution.havoc import _get_random_range_of_length

    # testing edge cases of length values
    with pytest.raises(ValueError):
        _get_random_range_of_length(bytearray([0] * INPUT_SIZE), 0)
    with pytest.raises(ValueError):
        _get_random_range_of_length(bytearray([0] * INPUT_SIZE), INPUT_SIZE + 1)

    # testing edge cases of input_ size
    with pytest.raises(ValueError):
        _get_random_range_of_length(bytearray([]), 1)

    input = bytearray([0] * INPUT_SIZE)
    for _ in range(ITERS):
        for length in [1, INPUT_SIZE]:
            start, end = _get_random_range_of_length(input, length)
            assert end - start == length
            assert start >= 0
            assert end <= INPUT_SIZE


def test_get_random_range_with_max_length():
    from fandango.evolution.havoc import _get_random_range_with_max_length

    # testing edge cases of length values
    with pytest.raises(ValueError):
        _get_random_range_with_max_length(bytearray([0] * INPUT_SIZE), 0)
    with pytest.raises(ValueError):
        _get_random_range_with_max_length(bytearray([0] * INPUT_SIZE), INPUT_SIZE + 1)

    # testing edge cases of input_ size
    with pytest.raises(ValueError):
        _get_random_range_with_max_length(bytearray([]), 1)

    input = bytearray([0] * INPUT_SIZE)
    for _ in range(1, ITERS):
        for max_length in [1, INPUT_SIZE]:
            start, end = _get_random_range_with_max_length(input, max_length)
            assert start < end
            assert start >= 0
            assert end <= INPUT_SIZE
            assert end - start <= max_length
            assert end - start >= 1


@pytest.mark.parametrize("mutation", havoc.havoc_mutations())
@pytest.mark.parametrize("input_size", [0, 1, 2, 255, 256])
def test_return_value(mutation, input_size):
    for _ in range(ITERS):
        input = bytearray(range(input_size))
        input_copy = input.copy()
        expected = mutation.mutate(input)
        actual = input != input_copy
        assert expected == actual, (
            f"Mutator {mutation} reported to {'not ' if not expected else ''}have mutated, "
            f"but the input was {'not ' if not actual else ''}mutated.\n"
            f"Original input: {input_copy}\nMutated input: {input}"
        )


@pytest.mark.parametrize("mutation", havoc.havoc_mutations())
@pytest.mark.parametrize("input_size", [0, 1, 2, 255, 256])
def test_size_changes(mutation, input_size):
    should_change_size = mutation.__class__.__name__ in [
        "BytesDeleteMutation",
        "BytesExpandMutation",
        "BytesInsertMutation",
        "BytesRandInsertMutation",
        "BytesInsertCopyMutation",
    ]
    for _ in range(ITERS):
        input = bytearray(range(input_size))
        expected_size = len(input)
        has_mutated = mutation.mutate(input)
        if has_mutated:
            assert (len(input) != expected_size) == should_change_size, (
                f"{mutation} mutated and should {'not ' if not should_change_size else ''}"
                f"change size but has {'not ' if should_change_size else ''}. "
                f"Input length: {len(input)}, expected size: {expected_size}"
            )
        else:
            assert (
                len(input) == expected_size
            ), f"{mutation} did not mutate but changed size"


def test_bit_flip_mutation():
    from fandango.evolution.havoc import BitFlipMutation

    mutation = BitFlipMutation()
    for _ in range(ITERS):
        input = bytearray([0] * INPUT_SIZE)
        assert mutation.mutate(input)
        c = Counter(input)
        assert c[0] == INPUT_SIZE - 1
        assert len(c.keys()) == 2
        other_key = [k for k in c.keys() if k != 0][0]
        assert bin(other_key).lstrip("0b").rstrip("0") == "1"


def test_byte_flip_mutation():
    from fandango.evolution.havoc import ByteFlipMutation

    mutation = ByteFlipMutation()
    for _ in range(ITERS):
        input = bytearray([0] * INPUT_SIZE)
        assert mutation.mutate(input)
        c = Counter(input)
        assert c[0] == INPUT_SIZE - 1
        assert c[0xFF] == 1


def test_byte_inc_mutation():
    from fandango.evolution.havoc import ByteIncMutation

    mutation = ByteIncMutation()

    for _ in range(ITERS):
        for value in [0, 1, 255]:
            input = bytearray([value] * INPUT_SIZE)
            assert mutation.mutate(input)
            c = Counter(input)
            assert c[value] == INPUT_SIZE - 1
            assert c[(value + 1) % 256] == 1


def test_byte_dec_mutation():
    from fandango.evolution.havoc import ByteDecMutation

    mutation = ByteDecMutation()
    for _ in range(ITERS):
        for value in [0, 1, 255]:
            input = bytearray([value] * INPUT_SIZE)
            assert mutation.mutate(input)
            c = Counter(input)
            assert c[value] == INPUT_SIZE - 1
            assert c[(value - 1) % 256] == 1


def test_byte_neg_mutation():
    from fandango.evolution.havoc import ByteNegMutation

    mutation = ByteNegMutation()
    for _ in range(ITERS):
        input = bytearray([1] * INPUT_SIZE)
        assert mutation.mutate(input)
        c = Counter(input)
        assert c[1] == INPUT_SIZE - 1
        assert c[255] == 1, f"Found unexpected count: {str(c)} in input: {str(input)}"

        input = bytearray(range(256))
        # 128 maps back to 128, so it's possible there is no mutation
        if mutation.mutate(input):
            found_mutation = False
            for i in range(len(input)):
                if i != input[i]:
                    assert not found_mutation, "Found multiple mutations"
                    found_mutation = True
                    assert input[i] + i == 256
            assert found_mutation, "No mutation found yet reported to be mutated"


def test_byte_rand_mutation():
    from fandango.evolution.havoc import ByteRandMutation

    mutation = ByteRandMutation()
    for _ in range(ITERS):
        for i in [0, 1, 255]:
            input = bytearray([i] * INPUT_SIZE)
            mutated = mutation.mutate(input)
            assert mutated == (
                input != (bytearray([i] * INPUT_SIZE))
            ), f"Found unexpected mutated input: {str(input)}"


@pytest.mark.parametrize(
    "mode",
    [(havoc.ArithmeticOperation.ADD, 0xFF), (havoc.ArithmeticOperation.SUB, 0x00)],
)
@pytest.mark.parametrize("num_bytes", [1, 2, 4, 8])
@pytest.mark.parametrize("little_endian", [True, False])
def test_multi_byte_arithmetic_mutation(mode, num_bytes, little_endian):
    from fandango.evolution.havoc import MultiByteArithmeticMutation

    mutation = MultiByteArithmeticMutation(num_bytes, mode[0], little_endian)
    for _ in range(ITERS):
        input = bytearray([mode[1]] * INPUT_SIZE)
        has_mutated = mutation.mutate(input)
        if has_mutated:
            assert input != bytearray([mode[1]] * INPUT_SIZE)
            # overflow should lead to all values being changed
            assert num_bytes == len([i for i in input if i != mode[1]])
        else:
            assert input == bytearray([mode[1]] * INPUT_SIZE)


@pytest.mark.parametrize("num_bytes", [1, 2, 4])
def test_multi_byte_interesting_mutation(num_bytes):
    from fandango.evolution.havoc import MultiByteInterestingMutation

    UNINTERESTING_BYTE = ord("A")
    mutation = MultiByteInterestingMutation(num_bytes)
    for _ in range(ITERS):
        input = bytearray([UNINTERESTING_BYTE] * INPUT_SIZE)
        assert mutation.mutate(input)
        c = Counter(input)
        assert c[UNINTERESTING_BYTE] == INPUT_SIZE - num_bytes
        other_values = [v for (k, v) in c.items() if k != UNINTERESTING_BYTE]
        assert sum(other_values) == num_bytes


def test_bytes_delete_mutation():
    from fandango.evolution.havoc import BytesDeleteMutation

    mutation = BytesDeleteMutation()
    for _ in range(ITERS):
        input = bytearray(range(256))
        assert mutation.mutate(input)
        assert len(input) < 256
        c = Counter(adjacent_diff(input))
        other_keys = [k for k in c.keys() if k != 1]

        match len(other_keys):
            case 0:
                # deletion from the extremes
                mutated_from_start = input[0] != 0
                mutated_from_end = input[-1] != 255
                assert mutated_from_start ^ mutated_from_end
            case 1:
                # deletion from the middle
                assert len(input) + other_keys[0] == 257  # diff => prior length + 1
            case _:
                # should never happen
                assert False


def test_bytes_expand_mutation():
    from fandango.evolution.havoc import BytesExpandMutation

    mutation = BytesExpandMutation()
    for _ in range(ITERS):
        input = bytearray(range(256))
        assert mutation.mutate(input)
        assert input[256:] in input[:256]


def test_bytes_insert_mutation():
    from fandango.evolution.havoc import BytesInsertMutation

    mutation = BytesInsertMutation()
    for _ in range(ITERS):
        input = bytearray(range(256))
        assert mutation.mutate(input)
        start = -1
        stop = -1
        jumps = []
        for i in range(1, len(input)):
            if input[i] == input[i - 1]:
                if start == -1:
                    start = i
                elif len([k for k in input[start:i] if k != input[i]]) == 0:
                    # if all values between start and i are the same, we are in the repetition
                    stop = i
                else:
                    assert (
                        False
                    ), f"Found multiple repetitions after index {i} in input {input}. start: {start}, stop: {stop}"
            elif input[i] != input[i - 1] + 1:
                jumps.append(i)
        if len(jumps) > 2:
            assert False, f"Found unexpected number of jumps: {jumps} in input {input}"

        input = bytearray([0] * INPUT_SIZE)
        assert mutation.mutate(input)
        assert len(input) > INPUT_SIZE
        assert len(Counter(input)) == 1


def test_bytes_rand_insert_mutation():
    from fandango.evolution.havoc import BytesRandInsertMutation

    mutation = BytesRandInsertMutation()
    for _ in range(ITERS):
        input = bytearray(range(256))
        assert mutation.mutate(input)
        start = -1
        stop = -1
        jumps = []
        for i in range(1, len(input)):
            if input[i] == input[i - 1]:
                if start == -1:
                    start = i
                elif len([k for k in input[start:i] if k != input[i]]) == 0:
                    # if all values between start and i are the same, we are in the repetition
                    stop = i
                else:
                    assert (
                        False
                    ), f"Found multiple repetitions after index {i} in input {input}. start: {start}, stop: {stop}"
            elif input[i] != input[i - 1] + 1:
                jumps.append(i)
        if len(jumps) > 2:
            assert False, f"Found unexpected number of jumps: {jumps} in input {input}"

        input = bytearray([0] * INPUT_SIZE)
        assert mutation.mutate(input)
        assert len(input) > INPUT_SIZE
        c = Counter(input)
        other_keys = [k for k in c.keys() if k != 0]
        match len(other_keys):
            case 0:
                pass  # chose 0 again
            case 1:
                indices = [i for i, v in enumerate(input) if v == other_keys[0]]
                # ensure all other values are consecutive
                assert (
                    len(
                        [
                            v
                            for i, v in enumerate(indices[:-1])
                            if v + 1 != indices[i + 1]
                        ]
                    )
                    == 0
                )
            case _:
                assert False


def test_bytes_set_mutation():
    from fandango.evolution.havoc import BytesSetMutation

    mutation = BytesSetMutation()
    for _ in range(ITERS):
        input = bytearray(range(256))
        has_mutated = mutation.mutate(input)
        if not has_mutated:
            assert input == bytearray(range(256))
            return
        start = -1
        stop = -1
        jumps = []
        for i in range(1, len(input)):
            if input[i] == input[i - 1]:
                if start == -1:
                    start = i
                elif len([k for k in input[start:i] if k != input[i]]) == 0:
                    # if all values between start and i are the same, we are in the repetition
                    stop = i
                else:
                    assert (
                        False
                    ), f"Found multiple repetitions after index {i} in input {input}. start: {start}, stop: {stop}"
            elif input[i] != input[i - 1] + 1:
                jumps.append(i)
        if len(jumps) > 2:
            assert False, f"Found unexpected number of jumps: {jumps} in input {input}"

        input = bytearray([0] * INPUT_SIZE)
        mutation.mutate(input)
        assert len(Counter(input)) == 1


def test_bytes_rand_set_mutation():
    from fandango.evolution.havoc import BytesRandSetMutation

    mutation = BytesRandSetMutation()
    for _ in range(ITERS):
        input = bytearray(range(256))
        if not mutation.mutate(input):
            assert input == bytearray(range(256))
            return
        start = -1
        stop = -1
        jumps = []
        for i in range(1, len(input)):
            if input[i] == input[i - 1]:
                if start == -1:
                    start = i
                elif len([k for k in input[start:i] if k != input[i]]) == 0:
                    # if all values between start and i are the same, we are in the repetition
                    stop = i
                else:
                    assert (
                        False
                    ), f"Found multiple repetitions after index {i} in input {input}. start: {start}, stop: {stop}"
            elif input[i] != input[i - 1] + 1:
                jumps.append(i)
        if len(jumps) > 2:
            assert False, f"Found unexpected number of jumps: {jumps} in input {input}"

        input = bytearray([0] * INPUT_SIZE)
        if not mutation.mutate(input):
            assert input == bytearray([0] * INPUT_SIZE)
            return
        c = Counter(input)
        other_keys = [k for k in c.keys() if k != 0]
        match len(other_keys):
            case 0:
                pass  # chose 0 again
            case 1:
                indices = [i for i, v in enumerate(input) if v == other_keys[0]]
                # ensure all other values are consecutive
                assert (
                    len(
                        [
                            v
                            for i, v in enumerate(indices[:-1])
                            if v + 1 != indices[i + 1]
                        ]
                    )
                    == 0
                )
            case _:
                assert False


def test_bytes_copy_mutation():
    from fandango.evolution.havoc import BytesCopyMutation

    mutation = BytesCopyMutation()
    for _ in range(ITERS):
        input = bytearray(range(256))
        if not mutation.mutate(input):
            assert input == bytearray(range(256))
            return
        found_duplicate_slice = False
        for start in range(len(input)):
            for end in range(start + 1, len(input)):
                if input[start:end] in input[start + 1 :]:
                    found_duplicate_slice = True
                    break
            if found_duplicate_slice:
                break
        assert found_duplicate_slice, f"Did not find duplicate slice in input {input}"


def test_bytes_insert_copy_mutation():
    from fandango.evolution.havoc import BytesInsertCopyMutation

    mutation = BytesInsertCopyMutation()
    for _ in range(ITERS):
        input = bytearray(range(256))
        input_copy = input.copy()
        assert mutation.mutate(input)
        diffs = []
        for i in range(1, len(input)):
            if input[i] != input[i - 1] + 1:
                diffs.append(i)
        if len(diffs) == 1:
            # insertion to the extremes
            i = diffs[0]
            if input[:i] == input_copy:  # insertion to the end
                assert input[i:] in input[:i]
            elif input[i:] == input_copy:  # insertion to the start
                assert input[:i] in input[i:]
        elif len(diffs) == 2:
            # insertion to the middle
            i, j = diffs
            temp_reconstructed = input[:i] + input[j:]
            assert (
                input[i:j] in temp_reconstructed
            ), f"input: {input}\ninserted slice (from {i} to {j}): {input[i:j]}"
        else:
            assert False, f"Found unexpected number of diffs: {diffs}"


def test_bytes_swap_mutation():
    from fandango.evolution.havoc import BytesSwapMutation

    mutation = BytesSwapMutation()
    for _ in range(ITERS):
        input = bytearray(range(256))
        input_copy = input.copy()
        if not mutation.mutate(input):
            assert input == input_copy
            return

        assert input != input_copy, "Input is not mutated"
        assert (
            bytearray(sorted(input)) == input_copy
        ), "Input does not contain all elements anymore"



import enum
from thefuzz import process as thefuzz_process


class FuzzingMode(enum.Enum):
    COMPLETE = 0
    IO = 1


def closest_match(word, candidates):
    """
    `word` raises a syntax error;
    return alternate suggestion for `word` from `candidates`
    """
    return thefuzz_process.extractOne(word, candidates)[0]


class ParsingMode(enum.Enum):
    COMPLETE = 0
    INCOMPLETE = 1



from fandango.language.symbols import Symbol, SymbolType
from fandango.logger import LOGGER


class Slice(Symbol):
    def __init__(self) -> None:
        super().__init__("", SymbolType.SLICE)

    def __hash__(self) -> int:
        return hash(self._type)

    def format_as_spec(self) -> str:
        LOGGER.warning(
            "Slice.format_as_spec() called. If this does not behave as expected, please report this as a bug."
        )
        return str(self._value)



#!/usr/bin/env python3
# Generate a Python parser for the C++ parser

from speedy_antlr_tool import generate

generate(
    py_parser_path="parser/FandangoParser.py",
    cpp_output_dir="cpp_parser",
    entry_rule_names=["fandango"],
)



import logging
from typing import Optional

from pygls.lsp.server import LanguageServer
import lsprotocol.types as lsp
from antlr4 import InputStream, Token

from fandango.language.parser.FandangoLexer import FandangoLexer
from fandango.language.server.semantic_tokens import (
    SemanticTokenModifiers,
    SemanticTokenTypes,
)

logger = logging.getLogger(__name__)


def map_to_ranges(references: list[Token], uri: str):
    def mapper(t: Token):
        assert isinstance(t.line, int), f"Token {t} has no line"
        assert isinstance(t.column, int), f"Token {t} has no column"
        return lsp.Range(
            start=lsp.Position(line=t.line - 1, character=t.column),
            end=lsp.Position(line=t.line - 1, character=t.column + len(t.text)),
        )

    return list(map(mapper, references))


def map_to_locations(references: list[Token], uri: str):
    return [lsp.Location(uri, range) for range in map_to_ranges(references, uri)]


class FileAssets:
    def __init__(self, lexer: FandangoLexer):
        self.tokens: list[Token] = lexer.getAllTokens()


class FandangoLanguageServer(LanguageServer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.files: dict[str, FileAssets] = {}

    def parse(self, uri: str):
        document = self.workspace.get_text_document(uri)
        document_text = document.source
        lexer = FandangoLexer(InputStream(document_text))
        self.files[document.uri] = FileAssets(lexer)

    def get_file_assets(self, document: lsp.TextDocumentIdentifier):
        return self.files[document.uri]

    def get_nonterminals(self, document: lsp.TextDocumentIdentifier):
        tokens = self.get_file_assets(document).tokens
        nonterminals = [t for t in tokens if t.type == FandangoLexer.NAME]
        return nonterminals

    def get_references(
        self,
        document: lsp.TextDocumentIdentifier,
        position: lsp.Position,
    ):
        non_terminals = self.get_nonterminals(document)
        current_token = [
            t
            for t in non_terminals
            if t.line == position.line + 1
            and t.column <= position.character
            and t.column + len(t.text) >= position.character
        ]

        references = [
            t for t in non_terminals if t.text in [ct.text for ct in current_token]
        ]

        return references


server = FandangoLanguageServer("fandango-language-server", "v0.1")


@server.feature(lsp.TEXT_DOCUMENT_DID_OPEN)
def did_open(ls: FandangoLanguageServer, params: lsp.DidOpenTextDocumentParams):
    """Parse each document when it is opened"""
    ls.parse(params.text_document.uri)


@server.feature(lsp.TEXT_DOCUMENT_DID_CHANGE)
def did_change(ls: FandangoLanguageServer, params: lsp.DidChangeTextDocumentParams):
    """Parse each document when it is changed"""
    ls.parse(params.text_document.uri)


# COMPLETION
@server.feature(
    lsp.TEXT_DOCUMENT_COMPLETION,
    lsp.CompletionOptions(trigger_characters=["<"], all_commit_characters=[">"]),
)
def completions(params: Optional[lsp.CompletionParams] = None) -> lsp.CompletionList:
    """Returns completion items."""
    if params is None:
        raise ValueError("params must not be None")

    texts = set(t.text for t in server.get_nonterminals(params.text_document))
    items = [lsp.CompletionItem(label=f"<{t}>", insert_text=f"{t}>") for t in texts]

    return lsp.CompletionList(is_incomplete=False, items=items)


@server.feature(lsp.TEXT_DOCUMENT_REFERENCES)
def find_references(server: FandangoLanguageServer, params: lsp.ReferenceParams):
    """Find references of an object."""

    return map_to_locations(
        server.get_references(params.text_document, params.position),
        params.text_document.uri,
    )


@server.feature(lsp.TEXT_DOCUMENT_DEFINITION)
def goto_definition(server: FandangoLanguageServer, params: lsp.DefinitionParams):
    """Jump to an object's definition."""

    tokens = server.get_file_assets(params.text_document).tokens
    references = server.get_references(params.text_document, params.position)

    for r in references:
        tokens_i = tokens.index(r)
        is_definition = False
        # continue until either newline (is not definition) or grammar assign (is definition)
        for i in range(tokens_i, len(tokens)):
            if tokens[i].type == FandangoLexer.NEWLINE:
                break
            elif tokens[i].type == FandangoLexer.GRAMMAR_ASSIGN:
                is_definition = True
                break

        if is_definition:
            return map_to_locations(
                [tokens[tokens_i]],
                params.text_document.uri,
            )[0]


@server.feature(lsp.TEXT_DOCUMENT_RENAME)
def rename(ls: FandangoLanguageServer, params: lsp.RenameParams):
    """Rename the symbol at the given position."""

    references = ls.get_references(params.text_document, params.position)
    edits = [
        lsp.TextEdit(range=range, new_text=params.new_name)
        for range in map_to_ranges(references, params.text_document.uri)
    ]

    return lsp.WorkspaceEdit(changes={params.text_document.uri: edits})


@server.feature(lsp.TEXT_DOCUMENT_PREPARE_RENAME)
def prepare_rename(ls: FandangoLanguageServer, params: lsp.PrepareRenameParams):
    """Called by the client to determine if renaming the symbol at the given location
    is a valid operation."""

    if len(ls.get_references(params.text_document, params.position)) == 0:
        return None

    return lsp.PrepareRenameDefaultBehavior(default_behavior=True)


@server.feature(
    lsp.TEXT_DOCUMENT_SEMANTIC_TOKENS_FULL,
    lsp.SemanticTokensLegend(
        token_types=SemanticTokenTypes.values(),
        token_modifiers=SemanticTokenModifiers.values(),
    ),
)
def semantic_tokens_full(ls: FandangoLanguageServer, params: lsp.SemanticTokensParams):
    tokens = ls.get_file_assets(params.text_document).tokens

    output = []
    prev_line = 0
    prev_col = 0

    for t in tokens:
        line = t.line - 1
        col = t.column

        delta_line = line - prev_line
        delta_start = col - prev_col if delta_line == 0 else col

        output.extend(
            [
                delta_line,
                delta_start,
                len(t.text),
                SemanticTokenTypes.from_token_as_number(t),
                SemanticTokenModifiers.from_token_as_number(t),
            ]
        )

        prev_line = line
        prev_col = col

    return lsp.SemanticTokens(data=output)


@server.feature(
    lsp.TEXT_DOCUMENT_CODE_ACTION,
    lsp.CodeActionOptions(code_action_kinds=[lsp.CodeActionKind.QuickFix]),
)
def code_actions(params: lsp.CodeActionParams):
    if (
        params.range.start.line != params.range.end.line
        or params.range.start.character != params.range.end.character
    ):
        return []  # ignore selections for now

    line = params.range.start.line
    column = params.range.start.character

    tokens = server.get_file_assets(params.text_document).tokens

    current_token = next(
        (
            t
            for t in tokens
            if t.line - 1 == line
            and t.column <= column
            and t.column + len(t.text) >= column
        ),
        None,
    )

    if current_token is None:
        return []

    defined_tokens = set()
    for i, t in enumerate(tokens):
        if t.type != FandangoLexer.GRAMMAR_ASSIGN:
            continue
        non_terminals = [nt for nt in tokens[i:0:-1] if nt.type == FandangoLexer.NAME]
        if len(non_terminals) == 0:
            continue
        defined_tokens.add(non_terminals[0].text)

    if current_token.text in defined_tokens:
        return []

    position = lsp.Position(line=current_token.line, character=0)
    range = lsp.Range(start=position, end=position)  # start == end means insert
    text_edit = lsp.TextEdit(
        range=range,
        new_text=f"<{current_token.text}> ::= # TODO: Implement\n",
    )

    return [
        lsp.CodeAction(
            title=f"Add definition for '{t.text}'",
            kind=lsp.CodeActionKind.QuickFix,
            edit=lsp.WorkspaceEdit(
                changes={params.text_document.uri: [text_edit]},
            ),
        )
    ]


if __name__ == "__main__":
    server.start_io()



import logging
import os
import sys
from typing import Any, IO

from fandango import Fandango
from fandango.cli.commands import COMMANDS, run
from fandango.cli.shell import shell_command
from fandango.cli.parser import get_parser
from fandango.logger import LOGGER


def main(
    *argv: str, stdout: IO[Any] | None = sys.stdout, stderr: IO[Any] | None = sys.stderr
) -> int:
    if "-O" in sys.argv:
        sys.argv.remove("-O")
        os.execl(sys.executable, sys.executable, "-O", *sys.argv)

    if stdout is not None:
        sys.stdout = stdout
    if stderr is not None:
        sys.stderr = stderr

    parser = get_parser(in_command_line=True)
    args = parser.parse_args(argv or sys.argv[1:])

    LOGGER.setLevel(os.getenv("FANDANGO_LOG_LEVEL", "WARNING"))  # Default

    if args.quiet and args.quiet == 1:
        LOGGER.setLevel(logging.WARNING)  # (Back to default)
    elif args.quiet and args.quiet > 1:
        LOGGER.setLevel(logging.ERROR)  # Even quieter
    elif args.verbose and args.verbose == 1:
        LOGGER.setLevel(logging.INFO)  # Give more info
    elif args.verbose and args.verbose > 1:
        LOGGER.setLevel(logging.DEBUG)  # Even more info

    # Set parsing method for .fan files
    Fandango.parser = args.parser

    if args.command in COMMANDS:
        # LOGGER.info(args.command)
        command = COMMANDS[args.command]
        last_status = run(command, args)
    elif args.command is None or args.command == "shell":
        last_status = run(shell_command, args)
    else:
        parser.print_usage()
        last_status = 2

    return last_status


if __name__ == "__main__":
    sys.exit(main())



from typing import TYPE_CHECKING, Optional
from collections.abc import Iterator, Sequence
from fandango.errors import FandangoValueError
from fandango.language.grammar.has_settings import HasSettings
from fandango.language.grammar.nodes.alternative import Alternative
from fandango.language.grammar.nodes.node import Node, NodeType
from fandango.language.symbols.non_terminal import NonTerminal
from fandango.language.tree import DerivationTree

if TYPE_CHECKING:
    import fandango


class NonTerminalNode(Node):
    def __init__(
        self,
        symbol: NonTerminal,
        grammar_settings: Sequence["HasSettings"],
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
    ):
        self.symbol = symbol
        self.sender = sender
        self.recipient = recipient
        super().__init__(NodeType.NON_TERMINAL, grammar_settings)

    def fuzz(
        self,
        parent: DerivationTree,
        grammar: "fandango.language.grammar.grammar.Grammar",
        max_nodes: int = 100,
        in_message: bool = False,
    ):
        if self.symbol not in grammar:
            raise FandangoValueError(f"Symbol {self.symbol} not found in grammar")

        # Gmutator mutation (4)
        if self.settings.get("non_terminal_use_other_rule"):
            other_nodes = [
                rhs
                for nt, rhs in grammar.rules.items()
                if nt.name() != self.symbol.name()
            ]
            print("other_nodes", [n.format_as_spec() for n in other_nodes])
            alt = Alternative(other_nodes, self._grammar_settings)
            return alt.fuzz(parent, grammar, max_nodes, in_message)

        dummy_current_tree = DerivationTree(self.symbol)
        parent.add_child(dummy_current_tree)

        if grammar.is_use_generator(dummy_current_tree):
            dependencies = grammar.generator_dependencies(self.symbol)
            for nt in dependencies:
                NonTerminalNode(nt, self._grammar_settings).fuzz(
                    dummy_current_tree, grammar, max_nodes - 1
                )
            parameters = dummy_current_tree.children
            for p in parameters:
                p._parent = None
            generated = grammar.generate(self.symbol, parameters)
            # Prevent children from being overwritten without executing generator
            for child in generated.children:
                child.set_all_read_only(True)

            generated.sender = self.sender
            generated.recipient = self.recipient
            parent.set_children(parent.children[:-1])
            parent.add_child(generated)
            return
        parent.set_children(parent.children[:-1])

        assign_sender = None
        assign_recipient = None
        if not in_message and self.sender is not None:
            assign_sender = self.sender
            assign_recipient = self.recipient
            in_message = True

        current_tree = DerivationTree(
            self.symbol,
            [],
            sender=assign_sender,
            recipient=assign_recipient,
            read_only=False,
        )
        parent.add_child(current_tree)
        grammar[self.symbol].fuzz(current_tree, grammar, max_nodes - 1, in_message)

    def accept(
        self,
        visitor: "fandango.language.grammar.node_visitors.node_visitor.NodeVisitor",
    ):
        return visitor.visitNonTerminalNode(self)

    def format_as_spec(self) -> str:
        if self.sender is not None:
            if self.recipient is None:
                return f"<{self.sender}:{(self.symbol.format_as_spec())[:-1]}>"
            else:
                return f"<{self.sender}:{self.recipient}:{(self.symbol.format_as_spec())[:-1]}>"
        else:
            return self.symbol.format_as_spec()

    def __eq__(self, other):
        return isinstance(other, NonTerminalNode) and self.symbol == other.symbol

    def __hash__(self):
        return hash(self.symbol)

    def msg_parties(self, *, include_recipients: bool = False) -> set[str]:
        parties: set[str] = super().msg_parties(include_recipients=include_recipients)
        if self.sender is not None:
            parties.add(self.sender)
            if self.recipient is not None and include_recipients:
                parties.add(self.recipient)
        return parties

    def descendents(
        self, grammar: "fandango.language.grammar.grammar.Grammar"
    ) -> Iterator["Node"]:
        yield grammar.rules[self.symbol]

    def in_parties(self, parties: list[str]) -> bool:
        return not self.sender or self.sender in parties



import argparse
import atexit
import os
import shlex
import sys
from typing import Any

from fandango.cli.commands import COMMANDS, help_command, run, version_command
from fandango.cli.complete import complete
from fandango.cli.parser import get_parser
from fandango.cli.utils import exec_single
from fandango.errors import FandangoError
from fandango.logger import LOGGER, print_exception

if "readline" not in globals():
    try:
        # Linux and Mac. This should do the trick.
        import gnureadline as readline  # type: ignore [import-not-found] # types not always available
    except Exception:
        pass

if "readline" not in globals():
    try:
        # Windows. This should do the trick.
        import pyreadline3 as readline  # type: ignore [import-not-found] # types not always available
    except Exception:
        pass

if "readline" not in globals():
    try:
        # Another Windows alternative
        import pyreadline as readline  # type: ignore [import-not-found] # types not always available
    except Exception:
        pass

if "readline" not in globals():
    try:
        # A Hail Mary Pass
        import readline
    except Exception:
        pass


MATCHES = []


def shell_command(args: argparse.Namespace) -> None:
    """Interactive mode"""

    PROMPT = "(fandango)"

    def _read_history() -> None:
        if "readline" not in globals():
            return

        histfile = os.path.join(os.path.expanduser("~"), ".fandango_history")
        try:
            readline.read_history_file(histfile)
            readline.set_history_length(1000)
        except FileNotFoundError:
            pass
        except Exception as e:
            LOGGER.warning(f"Could not read {histfile}: {e}")

        atexit.register(readline.write_history_file, histfile)

    def _complete(text: str, state: int) -> str | None:
        if "readline" not in globals():
            return None

        global MATCHES
        if state == 0:  # first trigger
            buffer = readline.get_line_buffer()[: readline.get_endidx()]
            MATCHES = complete(buffer)
        return MATCHES[state] if state < len(MATCHES) else None

    if sys.stdin.isatty():
        if "readline" in globals():
            _read_history()
            readline.set_completer_delims(" \t\n;")
            readline.set_completer(_complete)
            readline.parse_and_bind("tab: complete")

        version_command(argparse.Namespace())
        print("Type a command, 'help', 'copyright', 'version', or 'exit'.")

    while True:
        if sys.stdin.isatty():
            try:
                command_line = input(PROMPT + " ").lstrip()
            except KeyboardInterrupt:
                print("\nEnter a command, 'help', or 'exit'")
                continue
            except EOFError:
                break
        else:
            try:
                command_line = input().lstrip()
            except EOFError:
                break

        if command_line.startswith("!"):
            # Shell escape
            LOGGER.debug(command_line)
            if sys.stdin.isatty():
                os.system(command_line[1:])
            else:
                raise FandangoError(
                    "Shell escape (`!`) is only available in interactive mode"
                )
            continue

        if command_line.startswith("/"):
            # Python escape
            LOGGER.debug(command_line)
            if sys.stdin.isatty():
                try:
                    exec_single(command_line[1:].lstrip(), globals())
                except Exception as e:
                    print_exception(e)
            else:
                raise FandangoError(
                    "Python escape (`/`) is only available in interactive mode"
                )
            continue

        command: Any = None
        try:
            # hack to get this working for now — posix mode doesn't work with windows paths, non-posix mode doesn't do proper escaping
            posix = "win" not in sys.platform
            command = shlex.split(command_line, comments=True, posix=posix)
        except Exception as e:
            print_exception(e)
            continue

        if not command:
            continue

        if command[0].startswith("exit"):
            break

        parser = get_parser(in_command_line=False)
        try:
            args = parser.parse_args(command)
        except argparse.ArgumentError:
            parser.print_usage()
            continue
        except SystemExit:
            continue

        if args.command not in COMMANDS:
            parser.print_usage()
            continue

        LOGGER.debug(args.command + "(" + str(args) + ")")
        try:
            if args.command == "help":
                help_command(args, in_command_line=False)
            else:
                command = COMMANDS[args.command]
                run(command, args)
        except (SystemExit, KeyboardInterrupt):
            pass



import math
from typing import Any, Optional

from tdigest import TDigest as BaseTDigest

from fandango.constraints.fitness import (
    FailingTree,
    GeneticBase,
    ValueFitness,
)
from fandango.language.symbols import NonTerminal
from fandango.language.tree import DerivationTree
from fandango.logger import print_exception


class TDigest(BaseTDigest):
    def __init__(self, optimization_goal: str):
        super().__init__()
        self._min = None
        self._max = None
        self.contrast = 200.0
        if optimization_goal == "min":
            self.transform = self.amplify_near_0
        else:
            self.transform = self.amplify_near_1

    def update(self, x, w=1):
        super().update(x, w)
        if self._min is None or x < self._min:
            self._min = x
        if self._max is None or x > self._max:
            self._max = x

    def amplify_near_0(self, q):
        return 1 - math.exp(-self.contrast * q)

    def amplify_near_1(self, q):
        return math.exp(self.contrast * (q - 1))

    def score(self, x):
        if self._min is None or self._max is None:
            return 0
        if self._min == self._max:
            return self.transform(self.cdf(x))
        if x <= self._min:
            return 0
        if x >= self._max:
            return 1
        else:
            return self.transform(self.cdf(x))


class Value(GeneticBase):
    """
    Represents a value that can be used for fitness evaluation.
    In contrast to a constraint, a value is not calculated based on the constraints solved by a tree,
    but rather by a user-defined expression.
    """

    def __init__(self, expression: str, *args, **kwargs):
        """
        Initializes the value with the given expression.
        :param str expression: The expression to evaluate.
        :param args: Additional arguments.
        :param kwargs: Additional keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self.expression = expression
        self.cache: dict[int, ValueFitness] = dict()

    def fitness(
        self,
        tree: DerivationTree,
        scope: Optional[dict[NonTerminal, DerivationTree]] = None,
        population: Optional[list[DerivationTree]] = None,
        local_variables: Optional[dict[str, Any]] = None,
    ) -> ValueFitness:
        """
        Calculate the fitness of the tree based on the given expression.
        :param DerivationTree tree: The tree to evaluate.
        :param Optional[dict[NonTerminal, DerivationTree]] scope: The scope of the tree.
        :param Optional[list[DerivationTree]] population: The population of trees.
        :param Optional[dict[str, Any]] local_variables: Local variables to use in the evaluation.
        :return ValueFitness: The fitness of the tree.
        """
        tree_hash = self.get_hash(tree, scope, population, local_variables)
        # If the fitness has already been calculated, return the cached value
        if tree_hash in self.cache:
            return self.cache[tree_hash]
        # If the tree is None, the fitness is 0
        if tree is None:
            fitness = ValueFitness()
        else:
            trees = []
            values = []
            # Iterate over all combinations of the tree and the scope
            for combination in self.combinations(tree, scope, population):
                # Update the local variables to initialize the placeholders with the values of the combination
                local_vars = self.local_variables.copy()
                if local_variables:
                    local_vars.update(local_variables)
                local_vars.update(
                    {name: container.evaluate() for name, container in combination}
                )
                for _, container in combination:
                    for node in container.get_trees():
                        if node not in trees:
                            trees.append(node)
                try:
                    # Evaluate the expression
                    result = eval(self.expression, self.global_variables, local_vars)
                    values.append(result)
                except Exception as e:
                    print_exception(e, f"Evaluation failed: {self.expression}")
                    values.append(0)
            # Create the fitness object
            fitness = ValueFitness(
                values, failing_trees=[FailingTree(t, self) for t in trees]
            )
        # Cache the fitness
        self.cache[tree_hash] = fitness
        return fitness

    def get_symbols(self):
        """
        Get the placeholders of the constraint.
        """
        return self.searches.values()


class SoftValue(Value):
    """
    A `Value`, which is not mandatory, but aimed to be optimized.
    """

    def __init__(self, optimization_goal: str, expression: str, *args, **kwargs):
        super().__init__(expression, *args, **kwargs)
        assert optimization_goal in (
            "min",
            "max",
        ), f"Invalid SoftValue optimization goal {type!r}"
        self.optimization_goal = optimization_goal
        self.tdigest = TDigest(optimization_goal)

    def format_as_spec(self) -> str:
        representation = self.expression
        for identifier in self.searches:
            representation = representation.replace(
                identifier, self.searches[identifier].format_as_spec()
            )

        # noinspection PyUnreachableCode
        match self.optimization_goal:
            case "min":
                return f"minimizing {representation}"
            case "max":
                return f"maximizing {representation}"
            case _:
                return f"{self.optimization_goal} {representation}"



#!/usr/bin/env python3

# needs to be before any other imports
import os
from fandango.beartype import activate_beartype

if os.environ.get("FANDANGO_RUN_BEARTYPE", False):
    activate_beartype()

import importlib.metadata
import logging
import sys

from fandango.api import Fandango
from fandango.errors import (
    FandangoError,
    FandangoParseError,
    FandangoSyntaxError,
    FandangoValueError,
)
import fandango.language.tree
import fandango.language.parse

__all__ = [
    "FandangoError",
    "FandangoParseError",
    "FandangoSyntaxError",
    "FandangoValueError",
    "FandangoBase",
    "DerivationTree",
    "version",
    "homepage",
]

DISTRIBUTION_NAME = "fandango-fuzzer"


def version() -> str:
    """Return the Fandango version number"""
    return importlib.metadata.version(DISTRIBUTION_NAME)


def homepage() -> str:
    """Return the Fandango homepage"""
    metadata = importlib.metadata.metadata(DISTRIBUTION_NAME)
    return [
        e.split(",")[1].strip()
        for e in metadata.get_all("Project-URL", [])
        if e.startswith("homepage,")
    ].pop(0) or "the Fandango homepage"


DerivationTree = fandango.language.tree.DerivationTree
parse = fandango.language.parse.parse


if __name__ == "__main__":
    # Example Usage

    # Set the logging level (for debugging)
    logging_level = None
    if "-vv" in sys.argv:
        logging_level = logging.DEBUG
    elif "-v" in sys.argv:
        logging_level = logging.INFO

    # Read in a .fan spec (from a string)
    # We could also pass an (open) file or a list of files
    spec = """
        <my_start> ::= 'a' | 'b' | 'c'
        where str(<my_start>) != 'd'
    """
    fan = Fandango(spec, logging_level=logging_level, start_symbol="<my_start>")

    # Instantiate the spec into a population of derivation trees
    population = fan.fuzz(extra_constraints=["<my_start> != 'e'"], population_size=3)
    print("Fuzzed:", ", ".join(str(individual) for individual in population))

    # Parse a single input_ into a derivation tree
    trees = fan.parse("a")
    print("Parsed:", ", ".join(str(individual) for individual in trees))
